"""package querying interface

pquery is used to extract various kinds of information about either installed
or uninstalled packages. From an overall usage standpoint it is similar to
equery, but it can do things equery cannot do and is a bit more flexible.

What pquery does is select packages from one or more repos that match
a boolean combination of restrictions, then print selected information about
those packages. It is important to understand that the information printing and
repo selection options are almost completely separate from the
restriction options. The only exception to that is that restrictions on
contents automatically select the vdb (installed packages) repo, since
running them on source repos makes no sense.
"""

import errno
import os
from functools import partial

from snakeoil.cli import arghparse
from snakeoil.formatters import decorate_forced_wrapping
from snakeoil.osutils import sizeof_fmt
from snakeoil.sequences import iter_stable_unique

from ..ebuild import atom, conditionals
from ..fs import fs as fs_module
from ..repository import multiplex
from ..repository.util import get_raw_repos, get_virtual_repos
from ..restrictions import boolean, packages, values
from ..util import commandline
from ..util import packages as pkgutils
from ..util import parserestrict


class DataSourceRestriction(values.base):
    """Turn a data_source into a line iterator and apply a restriction."""

    def __init__(self, childrestriction, **kwargs):
        super().__init__(**kwargs)
        self.restriction = childrestriction

    def __str__(self):
        return f'DataSourceRestriction: {self.restriction} negate={self.negate}'

    def __repr__(self):
        if self.negate:
            string = '<%s restriction=%r negate @%#8x>'
        else:
            string = '<%s restriction=%r @%#8x>'
        return string % (self.__class__.__name__, self.restriction, id(self))

    def match(self, value):
        return self.restriction.match(iter(value.text_fileobj())) ^ self.negate

    __hash__ = object.__hash__


dep_attrs = ['bdepend', 'depend', 'rdepend', 'pdepend', 'idepend']
metadata_attrs = dep_attrs
dep_attrs += list(f'raw_{x}' for x in dep_attrs)
dep_formatted_attrs = dep_attrs + ['restrict']
dep_formatted_attrs = frozenset(dep_attrs + ['restrict'])
dep_attrs = tuple(sorted(dep_attrs))

metadata_attrs += [
    'defined_phases',
    'description',
    'eapi',
    'fetchables',
    'distfiles',
    'homepage',
    'inherited',
    'iuse',
    'keywords',
    'license',
    'properties',
    'required_use',
    'restrict',
    'slot',
    'subslot',
    'use',
]
metadata_attrs = tuple(sorted(metadata_attrs))

printable_attrs = tuple(dep_formatted_attrs) + metadata_attrs
printable_attrs += (
    'all',
    'alldepends',
    'allmetadata',
    'category',
    'cbuild',
    'chost',
    'ctarget',
    'environment',
    'files',
    'fullver',
    'longdescription',
    'maintainers',
    'package',
    'path',
    'raw_alldepends',
    'repo',
    'revision',
    'source_repository',
    'uris',
    'version',
)
printable_attrs = tuple(sorted(set(printable_attrs)))


def stringify_attr(config, pkg, attr):
    """Grab a package attr and convert it to a string."""
    # config is currently unused but may affect display in the future.
    if attr in ('files', 'uris'):
        data = get_pkg_attr(pkg, 'fetchables')
        if data is None:
            return 'MISSING'
        if attr == 'files':
            def _format(node):
                return node.filename
        else:
            def _format(node):
                return ' '.join(node.uri)
        return conditionals.stringify_boolean(data, _format)

    if attr == 'use':
        # Combine a list of all enabled (including irrelevant) and all
        # available flags into a "enabled -disabled" style string.
        use = set(get_pkg_attr(pkg, 'use', ()))
        iuse = get_pkg_attr(pkg, 'iuse_stripped', ())
        result = sorted(iuse & use) + sorted('-' + val for val in (iuse - use))
        return ' '.join(result)

    value = get_pkg_attr(pkg, attr)
    if value is None:
        return 'MISSING'

    if attr in ('iuse', 'properties', 'defined_phases', 'inherited'):
        return ' '.join(sorted(str(v) for v in value))
    if attr in ('maintainers', 'homepage'):
        return ' '.join(str(v) for v in value)
    if attr == 'longdescription':
        return str(value)
    if attr == 'keywords':
        return ' '.join(sorted(value, key=lambda x: x.lstrip("~")))
    if attr == 'distfiles':
        # collapse depsets for raw repo pkgs -- no USE flags are enabled
        if isinstance(value, conditionals.DepSet):
            value = value.evaluate_depset([])
        return ' '.join(value)
    if attr == 'environment':
        return value.text_fileobj().read()
    if attr == 'repo':
        return str(get_pkg_attr(value, 'repo_id', 'no repo id'))
    # hackish.
    return str(value)


def _default_formatter(out, node):
    out.write(node, autoline=False)
    return False


@decorate_forced_wrapping(False)
def format_depends(out, node, func=_default_formatter):
    """Pretty-print a depset to a formatter.

    :param out: formatter.
    :param node: a :obj:`conditionals.DepSet`.
    :param func: callable taking a formatter and a depset payload.
        If it can format its value in a single line it should do that
        without writing a newline and return C{False}.
        If it needs multiple lines it should first write a newline, not write
        a terminating newline, and return C{True}.
    :return: The same kind of boolean func should return.
    """
    # Do this first since if it is a DepSet it is also an
    # AndRestriction (DepSet subclasses that).
    if isinstance(node, conditionals.DepSet):
        if not node.restrictions:
            return False
        if len(node.restrictions) == 1:
            # Force a newline first.
            out.write()
            return _internal_format_depends(out, node.restrictions[0], func)
        out.write()
        for child in node.restrictions[:-1]:
            _internal_format_depends(out, child, func)
            out.write()
        _internal_format_depends(out, node.restrictions[-1], func)
        return True
    # weird..
    return _internal_format_depends(out, node, func)


def _internal_format_depends(out, node, func):
    prefix = None
    if isinstance(node, boolean.OrRestriction):
        prefix = '|| ('
        children = node.restrictions
    elif (isinstance(node, boolean.AndRestriction) and not
          isinstance(node, atom.atom)):
        prefix = '('
        children = node.restrictions
    elif isinstance(node, packages.Conditional):
        assert len(node.restriction.vals) == 1
        prefix = '%s%s? (' % (node.restriction.negate and '!' or '',
                              list(node.restriction.vals)[0])
        children = node.payload
    if prefix:
        children = list(children)
        if len(children) == 1:
            out.write(prefix, ' ', autoline=False)
            out.first_prefix.append('    ')
            newline = _internal_format_depends(out, children[0], func)
            out.first_prefix.pop()
            if newline:
                out.write()
                out.write(')', autoline=False)
                return True
            else:
                out.write(' )', autoline=False)
                return False
        else:
            out.write(prefix)
            out.first_prefix.append('    ')
            for child in children:
                _internal_format_depends(out, child, func)
                out.write()
            out.first_prefix.pop()
            out.write(')', autoline=False)
            return True
    else:
        return func(out, node)


def format_attr(config, out, pkg, attr):
    """Grab a package attr and print it through a formatter."""
    # config is currently unused but may affect display in the future.
    if attr in dep_formatted_attrs:
        data = get_pkg_attr(pkg, attr)
        if data is None:
            out.write('MISSING')
        else:
            out.first_prefix.append('        ')
            if config.highlight_dep:
                def _format(out, node):
                    for highlight in config.highlight_dep:
                        if highlight.intersects(node):
                            out.write(out.bold, out.fg('cyan'), node,
                                      autoline=False)
                            return
                    out.write(node, autoline=False)
                format_depends(out, data, _format)
            else:
                format_depends(out, data)
            out.first_prefix.pop()
            out.write()
    elif attr in ('files', 'uris'):
        data = get_pkg_attr(pkg, 'fetchables')
        if data is None:
            out.write('MISSING')
            return
        if attr == 'files':
            def _format(out, node):
                out.write(node.filename, autoline=False)
        else:
            def _format(out, node):
                if not node.uri:
                    return False
                if len(node.uri) == 1:
                    out.write(node.uri[0], autoline=False)
                    return False
                out.write('|| (')
                out.first_prefix.append('    ')
                for uri in node.uri:
                    out.write(uri)
                out.first_prefix.pop()
                out.write(')', autoline=False)
                return True
        out.first_prefix.append('        ')
        format_depends(out, data, _format)
        out.first_prefix.pop()
        out.write()
    else:
        out.write(stringify_attr(config, pkg, attr))


def print_package(options, out, err, pkg):
    """Print a package."""
    if options.verbosity > 0:
        green = out.fg('green')
        out.write(out.bold, green, ' * ', out.fg(), pkg.cpvstr)
        out.wrap = True
        out.later_prefix = ['                  ']
        for attr in options.attr:
            out.write(green, f'     {attr}: ', out.fg(), autoline=False)
            format_attr(options, out, pkg, attr)
        for revdep in options.print_revdep:
            for name in dep_attrs:
                depset = get_pkg_attr(pkg, name)
                find_cond = getattr(depset, 'find_cond_nodes', None)
                if find_cond is None:
                    out.write(
                        green, '     revdep: ', out.fg(), name, ' on ',
                        str(revdep))
                    continue
                for key, restricts in depset.find_cond_nodes(depset.restrictions, True):
                    if not restricts and key.intersects(revdep):
                        out.write(
                            green, '     revdep: ', out.fg(), name, ' on ',
                            autoline=False)
                        if key == revdep:
                            # this is never reached...
                            out.write(out.bold, str(revdep))
                        else:
                            out.write(
                                str(revdep), ' through dep ', out.bold,
                                str(key))
                for key, restricts in depset.node_conds.items():
                    if key.intersects(revdep):
                        out.write(
                            green, '     revdep: ', out.fg(), name, ' on ',
                            autoline=False)
                        if key == revdep:
                            out.write(
                                out.bold, str(revdep), out.reset,
                                autoline=False)
                        else:
                            out.write(
                                str(revdep), ' through dep ', out.bold,
                                str(key), out.reset, autoline=False)
                        out.write(' if USE matches one of:')
                        for r in restricts:
                            out.write('                  ', str(r))
        out.write()
        out.later_prefix = []
        out.wrap = False
    elif options.one_attr:
        if options.atom:
            out.write('=', autoline=False)
        if options.atom or options.cpv:
            out.write(pkg.cpvstr, autoline=False)
            if options.display_slot:
                out.write(':', pkg.slot, autoline=False)
            if options.display_repo:
                out.write('::', pkg.repo.repo_id, autoline=False)
            out.write('|', autoline=False)
        out.write(stringify_attr(options, pkg, options.one_attr))
    else:
        printed_something = False
        out.autoline = False
        if (not options.contents) or options.cpv:
            printed_something = True
            if options.atom:
                out.write('=')
            out.write(pkg.cpvstr)
            if options.display_slot:
                out.write(':', pkg.slot)
            if options.display_repo:
                out.write('::', pkg.repo.repo_id)
        for attr in options.attr:
            if printed_something:
                out.write(' ')
            printed_something = True
            attr_str = stringify_attr(options, pkg, attr)
            out.write(f'{attr}="{attr_str}"')
        for revdep in options.print_revdep:
            for name in dep_attrs:
                depset = get_pkg_attr(pkg, name)
                if getattr(depset, 'find_cond_nodes', None) is None:
                    # TODO maybe be smarter here? (this code is
                    # triggered by virtuals currently).
                    out.write(f' {name} on {revdep}')
                    continue
                for key, restricts in depset.find_cond_nodes(depset.restrictions, True):
                    if not restricts and key.intersects(revdep):
                        out.write(f' {name} on {revdep} through {key}')
                for key, restricts in depset.node_conds.items():
                    if key.intersects(revdep):
                        restricts = ' or '.join(map(str, restricts))
                        out.write(f' {name} on {revdep} through {key} if USE {restricts},')
        # If we printed anything at all print the newline now
        out.autoline = True
        if printed_something:
            out.write()

    if options.contents:
        color = {
            fs_module.fsDir: [out.bold, out.fg('blue')],
            fs_module.fsLink: [out.bold, out.fg('cyan')],
        }
        for obj in sorted(obj for obj in get_pkg_attr(pkg, 'contents', ())):
            if options.color:
                out.write(*(color.get(obj.__class__, []) + [obj] + [out.reset]))
            else:
                out.write(f'{obj!r}')

    if options.size:
        size = 0
        files = 0
        for location in (obj.location for obj in get_pkg_attr(pkg, 'contents', ())):
            files += 1
            size += os.lstat(location).st_size
        out.write(f'Total files: {files}')
        out.write(f'Total size: {sizeof_fmt(size)}')


def print_packages_noversion(options, out, err, pkgs):
    """Print a summary of all versions for a single package."""
    if options.verbosity > 0:
        green = out.fg('green')
        out.write(out.bold, green, ' * ', out.fg(), pkgs[0].key)
        out.wrap = True
        out.later_prefix = ['                  ']
        versions = ' '.join(pkg.fullver for pkg in sorted(pkgs))
        out.write(green, '     versions: ', out.fg(), versions)
        # If we are already matching on all repos we do not need to duplicate.
        if not options.all_repos:
            versions = sorted(
                pkg.fullver for repo in options.domain.installed_repos
                for pkg in repo.itermatch(pkgs[0].unversioned_atom))
            if versions:
                out.write(green, '     installed: ', out.fg(), ' '.join(versions))
        for attr in options.attr:
            out.write(green, f'     {attr}: ', out.fg(),
                      stringify_attr(options, pkgs[-1], attr))
        out.write()
        out.wrap = False
        out.later_prefix = []
    elif options.one_attr:
        if options.atom:
            out.write('=', autoline=False)
        if options.atom or options.cpv:
            out.write(pkgs[0].key, autoline=False)
            if options.display_slot:
                out.write(':', pkgs[0].slot, autoline=False)
            if options.display_repo:
                out.write('::', pkgs[0].repo.repo_id, autoline=False)
            out.write('|', autoline=False)
        out.write(stringify_attr(options, pkgs[-1], options.one_attr))
    else:
        out.autoline = False
        out.write(pkgs[0].key)
        if options.display_slot:
            out.write(':', pkgs[0].slot, autoline=False)
        if options.display_repo:
            out.write('::', pkgs[0].repo.repo_id, autoline=False)
        for attr in options.attr:
            attr_str = stringify_attr(options, pkgs[-1], attr)
            out.write(f' {attr}="{attr_str}"')
        out.autoline = True
        out.write()


# note the usage of priorities throughout this argparse setup;
# priority 0 (commandline sets this):
#  basically, sort the config first (additions/removals/etc),
# priority 30:
#   sort the repos
# priority 50:
#  sort the query args individually (potentially accessing the config) along
#  or lines for each (thus multiple revdep args are or'd together)
# priority 90:
#  finally generate a final query object, a boolean and of all previous
#  queries.
# priority 100:
#  default priority for DelayedValue; anything else is setup then.

argparser = commandline.ArgumentParser(
    domain=True, description=__doc__, script=(__file__, __name__))

repo_group = argparser.add_argument_group(
    'repository matching options',
    description='options controlling which repos to inspect')
repo_group.add_argument(
    '--raw', action='store_true', default=False,
    help="disable configuration filtering",
    docs="""
        Disable configuration filtering that forces raw dependencies to be
        used, rather than the dependencies rendered via your USE configuration.
        Primarily useful for people who need to look under the hood- ebuild
        devs, PM tool authors, etc. Note this option ignores --domain if is
        specified.
    """)
repo_group.add_argument(
    '--unfiltered', action='store_true', default=False,
    help="disable all license and visibility filtering",
    docs="""
        Disable all package filtering mechanisms such as ACCEPT_KEYWORDS,
        ACCEPT_LICENSE, and package.mask.
    """)
repo_group.add_argument(
    '--virtuals', action='store', choices=('only', 'disable'),
    help='only match virtuals or disable virtuals matching entirely',
    docs="""
        This option requires one of two arguments, either 'only' or 'disable',
        which causes only virtuals to be matched or disables virtuals matching
        entirely, respectively.

        By default, virtuals are included during matching.
    """)


class RawAwareStoreRepoObject(commandline.StoreRepoObject):
    """Custom implementation that is aware of the --raw and --unfiltered options."""

    def _get_sections(self, config, namespace):
        if namespace.raw:
            self.repo_key = 'repos_raw'
        elif namespace.unfiltered:
            self.repo_key = 'unfiltered_repos'
        else:
            self.repo_key = 'repos'
        return super()._get_sections(config, namespace)

repo_mux = repo_group.add_mutually_exclusive_group()
# TODO: update docs when binpkg/vdb repos are configured via repos.conf
repo_mux.add_argument(
    '-r', '--repo', action=RawAwareStoreRepoObject,
    priority=29, allow_external_repos=True,
    help='repo to search (default from domain if omitted)',
    docs="""
        Select the repo to search in for matches. This includes all the
        configured repos in repos.conf as well as the special keywords binpkg,
        provided, and vdb that search the configured binary package repo,
        package.provided, and installed packages, respectively.

        By default, all configured repos except the vdb will be searched when
        this option isn't specified.
    """)
repo_mux.add_argument(
    '-E', '--ebuild-repos', action='store_true',
    help='search all ebuild repos',
    docs="Search within all ebuild repos, all non-ebuild repos are skipped.")
repo_mux.add_argument(
    '-B', '--binary-repos', action='store_true',
    help='search all binary repos',
    docs="Search within all binary repos, all non-binary repos are skipped.")
repo_mux.add_argument(
    '-I', '--installed', action='store_true',
    help='search installed packages',
    docs="Search within installed packages (alias for '--repo vdb').")
repo_mux.add_argument(
    '-A', '--all-repos', action='store_true',
    help='search all repos',
    docs="Search all available repos including the vdb.")


@argparser.bind_delayed_default(30, 'repos')
def setup_repos(namespace, attr):
    # Get repo(s) to operate on.
    if namespace.repo:
        # The store repo machinery handles --raw and --unfiltered for
        # us, thus it being the first check.
        repos = [namespace.repo]
    elif (namespace.contents or namespace.size or namespace._owns or
            namespace._owns_re or namespace.installed):
        repos = namespace.domain.installed_repos
    elif namespace.unfiltered:
        if namespace.all_repos:
            repos = list(namespace.domain.installed_repos)
            repos.extend(namespace.domain.unfiltered_repos)
        elif namespace.ebuild_repos:
            repos = namespace.domain.ebuild_repos_raw
        elif namespace.binary_repos:
            repos = namespace.domain.binary_repos_raw
        else:
            repos = namespace.domain.unfiltered_repos
    elif namespace.all_repos:
        repos = namespace.domain.repos
    elif namespace.ebuild_repos:
        repos = namespace.domain.ebuild_repos
    elif namespace.binary_repos:
        repos = namespace.domain.binary_repos
    else:
        repos = namespace.domain.source_repos

    if namespace.raw or namespace.virtuals:
        repos = get_raw_repos(repos)
    if namespace.virtuals:
        repos = get_virtual_repos(
            repos, namespace.virtuals == 'only')
    setattr(namespace, attr, repos)

query = argparser.add_argument_group(
    'package matching options',
    docs="""
        Each option specifies a restriction packages must match. Specifying
        the same option twice means "or" unless stated otherwise. Specifying
        multiple types of restrictions means "and" unless stated otherwise.
    """)

# for queries, use add_query always; this has the bookkeeping
# necessary to ensure the sub-query gets bound into the
# finalized query
_query_items = []
def add_query(*args, **kwds):
    if 'dest' not in kwds:
        # auto-determine destination name from long option(s)
        dest = [x for x in args if x.startswith(argparser.prefix_chars * 2) and len(x) > 2]
        if not dest:
            raise ValueError(f"no valid options for query dest names: {', '.join(args)}")
        dest = dest[0].lstrip(argparser.prefix_chars)
        kwds['dest'] = dest.replace('-', '_')
    _query_items.append(kwds['dest'])
    kwds.setdefault('final_priority', 50)
    if kwds.get('action', None) == 'append':
        kwds.setdefault('default', [])
    commandline.make_query(query, *args, **kwds)

def bind_add_query(*args, **kwds):
    def f(functor):
        kwds[kwds.pop('bind', 'type')] = functor
        add_query(*args, **kwds)
        return functor
    return f

@bind_add_query(
    nargs='*', dest='matches', metavar='TARGET',
    bind='final_converter', type=None,
    help="extended atom matching of pkgs")
def matches_finalize(targets, namespace):
    repos = multiplex.tree(*namespace.repos)

    # If current working dir is in a repo, build a path restriction; otherwise
    # match everything.
    if not targets:
        cwd = os.getcwd()
        if cwd in repos:
            targets = [cwd]
        else:
            return []

    restrictions = []
    for target in targets:
        try:
            restrictions.append(parserestrict.parse_match(target))
        except parserestrict.ParseError as e:
            if os.path.exists(target):
                try:
                    restrictions.append(repos.path_restrict(target))
                except ValueError as e:
                    argparser.error(e)
            else:
                argparser.error(e)
    if restrictions:
        return packages.OrRestriction(*restrictions)
    return []

add_query(
    '--all', action='append_const',
    const=packages.AlwaysTrue, type=None,
    help='match all packages',
    docs="""
        Match all packages which is equivalent to "pquery *". Note that if no
        query options are specified, this option is enabled.
    """)
add_query(
    '--has-use', action='append',
    type=parserestrict.comma_separated_containment('iuse_stripped'),
    help='exact string match on a USE flag')
add_query(
    '--license', action='append',
    type=parserestrict.comma_separated_containment('license'),
    help='exact match on a license')

query.add_argument(
    '--revdep', nargs=1,
    action=arghparse.Expansion,
    subst=(('--restrict-revdep', '%(0)s'), ('--print-revdep', '%(0)s')),
    help='shorthand for --restrict-revdep atom --print-revdep atom',
    docs="""
        An alias for '--restrict-revdep atom --print-revdep atom', but note
        that --print-revdep is slow so use --restrict-revdep if you just need a
        list.
    """)

query.add_argument(
    '--revdep-pkgs', nargs=1,
    action=arghparse.Expansion,
    subst=(('--restrict-revdep-pkgs', '%(0)s'), ('--print-revdep', '%(0)s')),
    help='shorthand for --restrict-revdep-pkgs atom --print-revdep atom',
    docs="""
        An alias for '--restrict-revdep-pkgs atom --print-revdep atom', but
        note that --print-revdep is slow so use --restrict-revdep if you just
        need a list.
    """)

@bind_add_query(
    '--restrict-revdep', action='append', default=[],
    help='dependency on an atom')
def parse_revdep(value):
    """Value should be an atom, packages with deps intersecting that match."""
    try:
        targetatom = atom.atom(value)
    except atom.MalformedAtom as e:
        raise argparser.error(e)
    val_restrict = values.FlatteningRestriction(
        atom.atom,
        values.AnyMatch(values.FunctionRestriction(targetatom.intersects)))
    return packages.OrRestriction(*list(
        packages.PackageRestriction(dep, val_restrict)
        for dep in dep_attrs))

def _revdep_pkgs_match(pkgs, value):
    return any(value.match(pkg) for pkg in pkgs)

@bind_add_query(
    '--restrict-revdep-pkgs', action='append', type=atom.atom,
    default=[], bind='final_converter',
    help='dependency on pkgs that match a specific atom')
def revdep_pkgs_finalize(sequence, namespace):
    if not sequence:
        return []
    l = []
    for atom_inst in sequence:
        for repo in namespace.repos:
            l.extend(repo.itermatch(atom_inst))
    # have our pkgs; now build the restrict.
    any_restrict = values.AnyMatch(
        values.FunctionRestriction(partial(_revdep_pkgs_match, tuple(l))))
    r = values.FlatteningRestriction(atom.atom, any_restrict)
    return list(packages.PackageRestriction(dep, r)
                for dep in dep_attrs)

@bind_add_query(
    '-S', '--description', action='append',
    help='regexp search on description and longdescription')
def parse_description(value):
    """Value is used as a regexp matching description or longdescription."""
    matcher = values.StrRegex(value, case_sensitive=False)
    return packages.OrRestriction(*list(
        packages.PackageRestriction(attr, matcher)
        for attr in ('description', 'longdescription')))

@bind_add_query(
    '--eapi', action='append',
    help='match packages using a given EAPI')
def parse_eapi(value):
    """Value is matched against package EAPI versions."""
    return packages.PackageRestriction(
        'eapi',
        values.StrExactMatch(value))

@bind_add_query(
    '--owns', action='append',
    help='exact match on an owned file/dir')
def parse_owns(value):
    return packages.PackageRestriction(
        'contents',
        values.AnyMatch(values.GetAttrRestriction(
            'location', values.StrExactMatch(value))))

@bind_add_query(
    '--owns-re', action='append',
    help='like "owns" but using a regexp for matching')
def parse_ownsre(value):
    """Value is a regexp matched against the string form of an fs object.

    This means the object kind is prepended to the path the regexp has
    to match.
    """
    return packages.PackageRestriction(
        'contents',
        values.AnyMatch(values.GetAttrRestriction(
            'location', values.StrRegex(value))))

@bind_add_query(
    '--maintainer', action='append',
    help='comma-separated list of regexes to search for maintainers')
def parse_maintainer(value):
    """
    Case insensitive Regex match on the combined 'name <email>' bit of
    metadata.xml's maintainer data.
    """
    if value:
        return packages.PackageRestriction(
            'maintainers',
            values.AnyMatch(values.UnicodeConversion(
            values.StrRegex(value.lower(), case_sensitive=False))))
    else:
        # empty string matches packages without a maintainer
        return packages.PackageRestriction(
            'maintainers',
            values.EqualityMatch(()))

@bind_add_query(
    '--maintainer-name', action='append',
    help='comma-separated list of maintainer name regexes to search for')
def parse_maintainer_name(value):
    """
    Case insensitive Regex match on the name bit of metadata.xml's
    maintainer data.
    """
    return packages.PackageRestriction(
        'maintainers',
        values.AnyMatch(values.GetAttrRestriction(
            'name', values.StrRegex(value.lower(), case_sensitive=False))))

@bind_add_query(
    '--maintainer-email', action='append',
    help='comma-separated list of maintainer email regexes to search for')
def parse_maintainer_email(value):
    """
    Case insensitive Regex match on the email bit of metadata.xml's
    maintainer data.
    """
    return packages.PackageRestriction(
        'maintainers',
        values.AnyMatch(values.GetAttrRestriction(
            'email', values.StrRegex(value.lower(), case_sensitive=False))))

@bind_add_query(
    '--environment', action='append',
    help='regexp search in environment.bz2')
def parse_envmatch(value):
    """Apply a regexp to the environment."""
    return packages.PackageRestriction(
        'environment', DataSourceRestriction(values.AnyMatch(
            values.StrRegex(value))))

# note the type=str; this is to suppress the default
# fallback of using match parsing.
add_query(
    '--pkgset', action=commandline.StoreConfigObject,
    nargs=1, type=str, priority=35, config_type='pkgset',
    help='find packages that match the given package set (world for example)')

# add a fallback if no restrictions are specified.
_query_items.append('_fallback_all')
def _add_all_if_needed(namespace, attr):
    val = [packages.AlwaysTrue]
    for query_attr in _query_items:
        if getattr(namespace, f'_{query_attr}', None):
            val = None
            break
    setattr(namespace, attr, val)

@bind_add_query(
    '-u', '--upgrade', action='store_true',
    metavar=None, type=None, bind='final_converter',
    help='match installed packages without best slotted version')
def pkg_upgrade(_value, namespace):
    pkgs = []
    for pkg in namespace.domain.all_installed_repos:
        matches = sorted(namespace.domain.all_source_repos.match(pkg.slotted_atom))
        if matches and matches[-1] != pkg:
            pkgs.append(matches[-1].versioned_atom)
    return packages.OrRestriction(*pkgs)

argparser.set_defaults(
    _fallback_all=arghparse.DelayedValue(_add_all_if_needed, priority=89))
argparser.set_defaults(
    query=commandline.BooleanQuery(_query_items, klass_type='and', priority=90))

output = argparser.add_argument_group('output options')
output.add_argument(
    '-1', '--first', action='store_true',
    help='stop when first match is found')
output.add_argument(
    '-a', '--atom', action=arghparse.Expansion,
    subst=(('--cpv',),),
    help='print =cat/pkg-3 instead of cat/pkg-3.',
    docs="""
        Output valid package atoms, e.g. =cat/pkg-3 instead of cat/pkg-3.

        Note that this option implies --cpv and has no effect if used with
        --no-version.
    """)
output.add_argument(
    '--cpv', action='store_true',
    help='print the category/package-version',
    docs="""
        Display output in the format of 'category/package-version' which is
        done by default, this option forces the output format if another output
        option (such as --contents) alters it.
    """)
output.add_argument(
    '-R', action='store_true', dest='display_repo',
    help='print the repo of the package')
output.add_argument(
    '--slot', action='store_true', dest='display_slot',
    help='print the slot of the package')

output_mux = output.add_mutually_exclusive_group()
output_mux.add_argument(
    '-n', '--no-version', action='store_true',
    dest='noversion',
    help='collapse multiple matching versions together')
output_mux.add_argument(
    '--min', action='store_true',
    help='show only the lowest version for each package')
output_mux.add_argument(
    '--max', action='store_true',
    help='show only the highest version for each package')
del output_mux

output.add_argument(
    '--blame', action=arghparse.Expansion,
    subst=(("--attr", "maintainers"),),
    help='shorthand for --attr maintainers')
output.add_argument(
    '--size', action='store_true',
    help='display size of all files owned by the package')
output.add_argument(
    '--contents', action='store_true',
    help='list files owned by the package')
output.add_argument(
    '--highlight-dep', action='append',
    type=atom.atom, default=[],
    help='highlight dependencies matching this atom')
output.add_argument(
    '--print-revdep', action='append',
    type=atom.atom, default=[],
    help='print what condition(s) trigger a dep')

output.add_argument(
    '--attr', action='append', choices=printable_attrs,
    metavar='attribute', default=[],
    help="print this attribute's value (can be specified more than once)",
    docs=f"""
        Print the given attribute's value. This option can be specified
        multiple times.

        Valid attributes: {', '.join(printable_attrs)}
    """)
output.add_argument(
    '--force-attr', action='append', dest='attr',
    metavar='attribute', default=[],
    help='like --attr but accepts any string as '
         'attribute name instead of only explicitly '
         'supported names')
one_attr_mux = output.add_mutually_exclusive_group()
one_attr_mux.add_argument(
    '--one-attr', choices=printable_attrs,
    metavar='attribute',
    help="print one attribute, suppresses other output")
one_attr_mux.add_argument(
    '--force-one-attr',
    metavar='attribute',
    help='like --one-attr but accepts any string as '
         'attribute name instead of only explicitly '
         'supported names')
del one_attr_mux


def get_pkg_attr(pkg, attr, fallback=None):
    if attr[0:4] == 'raw_':
        pkg = getattr(pkg, '_raw_pkg', pkg)
        attr = attr[4:]
    return getattr(pkg, attr, fallback)


@argparser.bind_final_check
def _validate_args(parser, namespace):
    if namespace.noversion:
        if namespace.contents:
            parser.error('both --no-version and --contents does not make sense')
        if namespace.min or namespace.max:
            parser.error('--no-version with --min or --max does not make sense')
        if namespace.print_revdep:
            parser.error('--print-revdep with --no-version does not make sense')

    if namespace.one_attr and namespace.print_revdep:
        parser.error('--print-revdep with --force-one-attr or --one-attr does not make sense')

    def process_attrs(sequence):
        for attr in sequence:
            if attr == 'all':
                i = [x for x in printable_attrs if x != 'all']
            elif attr == 'allmetadata':
                i = process_attrs(metadata_attrs)
            elif attr == 'alldepends':
                i = dep_attrs
            elif attr == 'raw_alldepends':
                i = [f'raw_{x}' for x in dep_attrs]
            else:
                i = [attr]
            for attr in i:
                yield attr

    attrs = ['repo', 'description', 'homepage', 'license'] if namespace.verbosity > 0 else []
    attrs.extend(process_attrs(namespace.attr))

    # finally, uniquify the attrs.
    namespace.attr = list(iter_stable_unique(attrs))


@argparser.bind_main_func
def main(options, out, err):
    """Run a query."""
    if options.debug:
        for repo in options.repos:
            out.write(f'repo: {repo.repo_id}')
        out.write(f'restrict: {options.query}')
        out.write()

    if options.query is None:
        return 0
    for repo in options.repos:
        try:
            for pkgs in pkgutils.groupby_pkg(repo.itermatch(options.query, sorter=sorted)):
                pkgs = list(pkgs)
                if options.noversion:
                    print_packages_noversion(options, out, err, pkgs)
                elif options.min or options.max:
                    if options.min:
                        print_package(options, out, err, min(pkgs))
                    if options.max:
                        print_package(options, out, err, max(pkgs))
                else:
                    for pkg in pkgs:
                        print_package(options, out, err, pkg)
                        if options.first:
                            break
                if options.first:
                    break

        except KeyboardInterrupt:
            raise
        except Exception as e:
            if isinstance(e, IOError) and e.errno == errno.EPIPE:
                # swallow it; receiving end shutdown early.
                return
            # force a newline for error msg or traceback output
            err.write()
            raise



from pkgcore.config import basics, central, errors
from pkgcore.config.hint import configurable
from snakeoil.errors import walk_exception_chain
from snakeoil.test import TestCase


# A bunch of functions used from various tests below.
def repo(cache):
    return cache
@configurable({'content': 'ref:drawer', 'contents': 'refs:drawer'})
def drawer(content=None, contents=None):
    return content, contents

# The exception checks here also check if the str value of the
# exception is what we expect. This does not mean the wording of the
# error messages used here is strictly required. It just makes sure
# the error we get is the expected one and is useful. Please make sure
# you check for a sensible error message when more tests are added.

# A lot of the ConfigManager instances get object() as a remote type.
# This makes sure the types are not unnecessarily queried (since
# querying object() will blow up).

class RemoteSource:

    """Use this one for tests that do need the names but nothing more."""

    def __iter__(self):
        return iter(('remote',))

    def __getitem__(self, key):
        raise NotImplementedError()


def _str_exc(exc):
    return ":\n".join(str(x) for x in walk_exception_chain(exc))

class ConfigManagerTest(TestCase):

    def check_error(self, message, func, *args, **kwargs):
        """Like assertRaises but checks for the message string too."""
        klass = kwargs.pop('klass', errors.ConfigurationError)
        try:
            func(*args, **kwargs)
        except klass as e:
            self.assertEqual(
                message, _str_exc(e),
                f'\nGot:\n{_str_exc(e)!r}\nExpected:\n{message!r}\n')
        else:
            self.fail('no exception raised')

    def get_config_obj(self, manager, obj_type, obj_name):
        types = getattr(manager.objects, obj_type)
        return types[obj_name]

    def test_sections(self):
        manager = central.ConfigManager(
            [{'fooinst': basics.HardCodedConfigSection({'class': repo}),
              'barinst': basics.HardCodedConfigSection({'class': drawer}),
              }])
        self.assertEqual(['barinst', 'fooinst'], sorted(manager.sections()))
        self.assertEqual(list(manager.objects.drawer.keys()), ['barinst'])
        self.assertEqual(manager.objects.drawer, {'barinst': (None, None)})

    def test_contains(self):
        manager = central.ConfigManager(
            [{'spork': basics.HardCodedConfigSection({'class': drawer})}],
            [RemoteSource()])
        self.assertIn('spork', manager.objects.drawer)
        self.assertNotIn('foon', manager.objects.drawer)

    def test_no_class(self):
        manager = central.ConfigManager(
            [{'foo': basics.HardCodedConfigSection({})}])
        self.check_error(
            "Collapsing section named 'foo':\n"
            'no class specified',
            manager.collapse_named_section, 'foo')

    def test_missing_section_ref(self):
        manager = central.ConfigManager(
            [{'rsync repo': basics.HardCodedConfigSection({'class': repo}),
              }])
        self.check_error(
            "Collapsing section named 'rsync repo':\n"
            "type tests.config.test_central.repo needs settings for "
            "'cache'",
            self.get_config_obj, manager, 'repo', 'rsync repo')

    def test_unknown_type(self):
        manager = central.ConfigManager(
            [{'spork': basics.HardCodedConfigSection({'class': drawer,
                                                      'foon': None})}])
        self.check_error(
            "Collapsing section named 'spork':\n"
            "Type of 'foon' unknown",
            manager.collapse_named_section, 'spork')

    def test_missing_inherit_target(self):
        manager = central.ConfigManager(
            [{'myrepo': basics.HardCodedConfigSection({
                            'class': repo,
                            'inherit': ['baserepo'],
                            }),
              }], [RemoteSource()])
        self.check_error(
            "Collapsing section named 'myrepo':\n"
            "Inherit target 'baserepo' cannot be found",
            self.get_config_obj, manager, 'repo', 'myrepo')

    def test_inherit_unknown_type(self):
        manager = central.ConfigManager(
            [{'baserepo': basics.HardCodedConfigSection({
                            'cache': 'available',
                            }),
              'actual repo': basics.HardCodedConfigSection({
                            'class': drawer,
                            'inherit': ['baserepo'],
                            }),
              }])
        self.check_error(
            "Collapsing section named 'actual repo':\n"
            "Type of 'cache' unknown",
            self.get_config_obj, manager, 'repo', 'actual repo')

    def test_inherit(self):
        manager = central.ConfigManager(
            [{'baserepo': basics.HardCodedConfigSection({
                            'cache': 'available',
                            'inherit': ['unneeded'],
                            }),
              'unneeded': basics.HardCodedConfigSection({
                            'cache': 'unavailable'}),
              'actual repo': basics.HardCodedConfigSection({
                            'class': repo,
                            'inherit': ['baserepo'],
                            }),
              }])

        self.assertEqual('available', manager.objects.repo['actual repo'])

    def test_no_object_returned(self):
        def noop():
            """Do not do anything."""
        manager = central.ConfigManager(
            [{'myrepo': basics.HardCodedConfigSection({'class': noop}),
              }])
        self.check_error(
            "Failed instantiating section 'myrepo':\n"
            "'No object returned' instantiating "
            "tests.config.test_central.noop",
            manager.collapse_named_section('myrepo').instantiate)

    def test_not_callable(self):
        class myrepo:
            def __repr__(self):
                return 'useless'
        manager = central.ConfigManager(
            [{'myrepo': basics.HardCodedConfigSection({'class': myrepo()}),
              }])
        self.check_error(
            "Collapsing section named 'myrepo':\n"
            "Failed converting argument 'class' to callable:\n"
            "useless is not callable",
            self.get_config_obj, manager, 'myrepo', 'myrepo')

    def test_raises_instantiationerror(self):
        def myrepo():
            raise Exception('I raised')
        manager = central.ConfigManager(
            [{'myrepo': basics.HardCodedConfigSection({'class': myrepo}),
              }])
        self.check_error(
            "Failed instantiating section 'myrepo':\n"
            "Failed instantiating section 'myrepo': exception caught from 'tests.config.test_central.myrepo':\n"
            "I raised",
            self.get_config_obj, manager, 'myrepo', 'myrepo')

    def test_raises(self):
        def myrepo():
            raise ValueError('I raised')
        manager = central.ConfigManager(
            [{'myrepo': basics.HardCodedConfigSection({'class': myrepo})
              }])
        self.check_error(
            "Failed instantiating section 'myrepo':\n"
            "Failed instantiating section 'myrepo': exception caught from 'tests.config.test_central.myrepo':\n"
            "I raised",
            self.get_config_obj, manager, 'myrepo', 'myrepo')
        manager = central.ConfigManager(
            [{'myrepo': basics.HardCodedConfigSection({'class': myrepo})
              }], debug=True)
        self.check_error(
            "Failed instantiating section 'myrepo':\n"
            "Failed instantiating section 'myrepo': exception caught from 'tests.config.test_central.myrepo':\n"
            "I raised",
                self.get_config_obj, manager, 'myrepo', 'myrepo',
                klass=errors.ConfigurationError)

    def test_pargs(self):
        @configurable(types={'p': 'str', 'notp': 'str'},
                      positional=['p'], required=['p'])
        def myrepo(*args, **kwargs):
            return args, kwargs
        manager = central.ConfigManager(
            [{'myrepo': basics.HardCodedConfigSection({
                            'class': myrepo,
                            'p': 'pos',
                            'notp': 'notpos',
                            }),
              }])

        self.assertEqual(
            manager.objects.myrepo['myrepo'], (('pos',), {'notp': 'notpos'}))

    def test_autoexec(self):
        @configurable(typename='configsection')
        def autoloader():
            return {'spork': basics.HardCodedConfigSection({'class': repo,
                                                            'cache': 'test'})}

        manager = central.ConfigManager(
            [{'autoload-sub': basics.HardCodedConfigSection({
                            'class': autoloader,
                            })}])
        self.assertEqual(set(['autoload-sub', 'spork']), set(manager.sections()))
        self.assertEqual(['spork'], list(manager.objects.repo.keys()))
        self.assertEqual(
            'test',
            manager.collapse_named_section('spork').instantiate())

    def test_reload(self):
        mod_dict = {'class': repo, 'cache': 'test'}

        @configurable(typename='configsection')
        def autoloader():
            return {'spork': basics.HardCodedConfigSection(mod_dict)}

        manager = central.ConfigManager(
            [{'autoload-sub': basics.HardCodedConfigSection({
                            'class': autoloader})}])

        self.assertEqual(set(['autoload-sub', 'spork']), set(manager.sections()))
        self.assertEqual(['spork'], list(manager.objects.repo.keys()))
        collapsedspork = manager.collapse_named_section('spork')
        self.assertEqual('test', collapsedspork.instantiate())
        mod_dict['cache'] = 'modded'
        self.assertIdentical(collapsedspork,
                             manager.collapse_named_section('spork'))
        self.assertEqual('test', collapsedspork.instantiate())
        types = manager.types
        manager.reload()
        newspork = manager.collapse_named_section('spork')
        self.assertNotIdentical(collapsedspork, newspork)
        self.assertEqual(
            'modded', newspork.instantiate(),
            'it did not throw away the cached instance')
        self.assertNotIdentical(types, manager.types)

    def test_instantiate_default_ref(self):
        manager = central.ConfigManager(
            [{'spork': basics.HardCodedConfigSection({'class': drawer})}],
            )
        self.assertEqual(
            (None, None),
            manager.collapse_named_section('spork').instantiate())

    def test_allow_unknowns(self):
        @configurable(allow_unknowns=True)
        def myrepo(**kwargs):
            return kwargs

        manager = central.ConfigManager(
            [{'spork': basics.HardCodedConfigSection({
                            'class': myrepo, 'spork': 'foon'})}])

        self.assertEqual(
            {'spork': 'foon'},
            manager.collapse_named_section('spork').instantiate())

    def test_reinstantiate_after_raise(self):
        # The most likely bug this tests for is attempting to
        # reprocess already processed section_ref args.
        spork = object()
        @configurable({'thing': 'ref:spork'})
        def myrepo(thing):
            self.assertIdentical(thing, spork)
            raise errors.ComplexInstantiationError('I suck')
        @configurable(typename='spork')
        def spork_producer():
            return spork
        manager = central.ConfigManager(
            [{'spork': basics.HardCodedConfigSection({
                            'class': myrepo,
                            'thing': basics.HardCodedConfigSection({
                                    'class': spork_producer,
                                    }),
                            })}])
        spork = manager.collapse_named_section('spork')
        for i in range(3):
            self.check_error(
                "Failed instantiating section 'spork':\n"
                "Failed instantiating section 'spork': exception caught from 'tests.config.test_central.myrepo':\n"
                "'I suck', callable unset!",
                spork.instantiate)
        for i in range(3):
            self.check_error(
                "Failed instantiating section 'spork':\n"
                "Failed instantiating section 'spork': exception caught from 'tests.config.test_central.myrepo':\n"
                "'I suck', callable unset!",
                manager.collapse_named_section('spork').instantiate)

    def test_instantiation_caching(self):
        @configurable(typename='drawer')
        def myrepo():
            return object()

        manager = central.ConfigManager(
            [{'spork': basics.HardCodedConfigSection({'class': myrepo}),
              'drawer': basics.ConfigSectionFromStringDict({
                            'class': 'tests.config.test_central.drawer',
                            'content': 'spork',
                            }),
              }])

        config = manager.collapse_named_section('spork')
        self.assertIdentical(config.instantiate(), config.instantiate())
        self.assertIdentical(
            config.instantiate(),
            manager.collapse_named_section('drawer').instantiate()[0])

    def test_collapse_named_errors(self):
        manager = central.ConfigManager(
            [{'spork': basics.ConfigSectionFromStringDict({
                            'class': 'tests.config.test_central.drawer',
                            'content': 'ref'})}], [RemoteSource()])
        self.assertRaises(KeyError, self.get_config_obj, manager, 'repo', 'foon')
        self.check_error(
            "Collapsing section named 'spork':\n"
            "Failed collapsing section key 'content':\n"
            "no section called 'ref'",
            self.get_config_obj, manager, 'repo', 'spork')

    def test_recursive_autoload(self):
        @configurable(typename='configsection')
        def autoloader():
            return {'autoload-sub': basics.HardCodedConfigSection(
                    {'class': autoloader}),
                    'spork': basics.HardCodedConfigSection({'class': repo,
                                                            'cache': 'test'})}

        self.check_error(
            "New config is trying to modify existing section(s) 'autoload-sub' "
            "that was already instantiated.",
            central.ConfigManager,
            [{'autoload-sub': basics.HardCodedConfigSection({
                            'class': autoloader,
                            })}])

    def test_recursive_section_ref(self):
        manager = central.ConfigManager(
            [{'spork': basics.ConfigSectionFromStringDict({
                            'class': 'tests.config.test_central.drawer',
                            'content': 'foon'}),
              'foon': basics.ConfigSectionFromStringDict({
                            'class': 'tests.config.test_central.drawer',
                            'content': 'spork'}),
              'self': basics.ConfigSectionFromStringDict({
                            'class': 'tests.config.test_central.drawer',
                            'content': 'self'}),
              }])
        self.check_error(
            "Collapsing section named 'self':\n"
            "Failed collapsing section key 'content':\n"
            "Reference to 'self' is recursive",
            self.get_config_obj, manager, 'drawer', 'self')
        self.check_error(
            "Collapsing section named 'spork':\n"
            "Failed collapsing section key 'content':\n"
            "Collapsing section named 'foon':\n"
            "Failed collapsing section key 'content':\n"
            "Reference to 'spork' is recursive",
            self.get_config_obj, manager, 'drawer', 'spork')

    def test_recursive_inherit(self):
        manager = central.ConfigManager(
            [{'spork': basics.ConfigSectionFromStringDict({
                            'class': 'tests.config.test_central.drawer',
                            'inherit': 'foon'}),
              'foon': basics.ConfigSectionFromStringDict({
                            'class': 'tests.config.test_central.drawer',
                            'inherit': 'spork'}),
              }])
        self.check_error(
            "Collapsing section named 'spork':\n"
            "Inherit 'spork' is recursive",
            self.get_config_obj, manager, 'drawer', 'spork')

    def test_alias(self):
        def myspork():
            return object
        manager = central.ConfigManager(
            [{'spork': basics.HardCodedConfigSection({'class': myspork}),
              'foon': basics.section_alias('spork', 'myspork'),
              }])
        # This tests both the detected typename of foon and the caching.
        self.assertIdentical(manager.objects.myspork['spork'], manager.objects.myspork['foon'])

    def test_typecheck(self):
        @configurable({'myrepo': 'ref:repo'}, typename='repo')
        def reporef(myrepo=None):
            return myrepo
        @configurable({'myrepo': 'refs:repo'}, typename='repo')
        def reporefs(myrepo=None):
            return myrepo
        @configurable(typename='repo')
        def myrepo():
            return 'repo!'
        manager = central.ConfigManager([{
                    'myrepo': basics.HardCodedConfigSection({'class': myrepo}),
                    'drawer': basics.HardCodedConfigSection({'class': drawer}),
                    'right':  basics.AutoConfigSection({'class': reporef,
                                                        'myrepo': 'myrepo'}),
                    'wrong':  basics.AutoConfigSection({'class': reporef,
                                                        'myrepo': 'drawer'}),
                    }])
        self.check_error(
            "Collapsing section named 'wrong':\n"
            "Failed collapsing section key 'myrepo':\n"
            "reference 'drawer' should be of type 'repo', got 'drawer'",
            self.get_config_obj, manager, 'repo', 'wrong')
        self.assertEqual('repo!', manager.objects.repo['right'])

        manager = central.ConfigManager([{
                    'myrepo': basics.HardCodedConfigSection({'class': myrepo}),
                    'drawer': basics.HardCodedConfigSection({'class': drawer}),
                    'right':  basics.AutoConfigSection({'class': reporefs,
                                                        'myrepo': 'myrepo'}),
                    'wrong':  basics.AutoConfigSection({'class': reporefs,
                                                        'myrepo': 'drawer'}),
                    }])
        self.check_error(
            "Collapsing section named 'wrong':\n"
            "Failed collapsing section key 'myrepo':\n"
            "reference 'drawer' should be of type 'repo', got 'drawer'",
            self.get_config_obj, manager, 'repo', 'wrong')
        self.assertEqual(['repo!'], manager.objects.repo['right'])

    def test_default(self):
        manager = central.ConfigManager([{
                    'thing': basics.HardCodedConfigSection({'class': drawer,
                                                            'default': True}),
                    'bug': basics.HardCodedConfigSection({'class': None,
                                                          'inherit-only':True,
                                                          'default': True}),
                    'ignore': basics.HardCodedConfigSection({'class': drawer}),
                    }])
        self.assertEqual((None, None), manager.get_default('drawer'))
        self.assertTrue(manager.collapse_named_section('thing').default)

        manager = central.ConfigManager([{
                    'thing': basics.HardCodedConfigSection({'class': drawer,
                                                            'default': True}),
                    'thing2': basics.HardCodedConfigSection({'class': drawer,
                                                             'default': True}),
                    }])
        self.check_error(
            "type drawer incorrectly has multiple default sections: 'thing', 'thing2'",
            manager.get_default, 'drawer')

        manager = central.ConfigManager([])
        self.assertIdentical(None, manager.get_default('drawer'))

    def test_broken_default(self):
        def broken():
            raise errors.ComplexInstantiationError('broken')
        manager = central.ConfigManager([{
                    'thing': basics.HardCodedConfigSection({
                            'class': drawer, 'default': True,
                            'content': basics.HardCodedConfigSection({
                                    'class': 'spork'})}),
                    'thing2': basics.HardCodedConfigSection({
                            'class': broken, 'default': True})}])
        self.check_error(
            "Collapsing defaults for 'drawer':\n"
            "Collapsing section named 'thing':\n"
            "Failed collapsing section key 'content':\n"
            "Failed converting argument 'class' to callable:\n"
            "'spork' is not callable",
            manager.get_default, 'drawer')
        self.check_error(
            "Collapsing defaults for 'broken':\n"
            "Collapsing section named 'thing':\n"
            "Failed collapsing section key 'content':\n"
            "Failed converting argument 'class' to callable:\n"
            "'spork' is not callable",
            manager.get_default, 'broken')

    def test_instantiate_broken_ref(self):
        @configurable(typename='drawer')
        def broken():
            raise errors.ComplexInstantiationError('broken')
        manager = central.ConfigManager([{
                    'one': basics.HardCodedConfigSection({
                            'class': drawer,
                            'content': basics.HardCodedConfigSection({
                                    'class': broken})}),
                    'multi': basics.HardCodedConfigSection({
                            'class': drawer,
                            'contents': [basics.HardCodedConfigSection({
                                        'class': broken})]}),
                    }])
        self.check_error(
            "Failed instantiating section 'one':\n"
            "Instantiating reference 'content' pointing at None:\n"
            "Failed instantiating section None:\n"
            "Failed instantiating section None: exception caught from 'tests.config.test_central.broken':\n"
            "'broken', callable unset!",
            manager.collapse_named_section('one').instantiate)
        self.check_error(
            "Failed instantiating section 'multi':\n"
            "Instantiating reference 'contents' pointing at None:\n"
            "Failed instantiating section None:\n"
            "Failed instantiating section None: exception caught from 'tests.config.test_central.broken':\n"
            "'broken', callable unset!",
            manager.collapse_named_section('multi').instantiate)

    def test_autoload_instantiationerror(self):
        @configurable(typename='configsection')
        def broken():
            raise errors.ComplexInstantiationError('broken')
        self.check_error(
            "Failed loading autoload section 'autoload_broken':\n"
            "Failed instantiating section 'autoload_broken':\n"
            "Failed instantiating section 'autoload_broken': exception caught from 'tests.config.test_central.broken':\n"
            "'broken', callable unset!",
            central.ConfigManager, [{
                    'autoload_broken': basics.HardCodedConfigSection({
                            'class': broken})}])

    def test_autoload_uncollapsable(self):
        self.check_error(
            "Failed collapsing autoload section 'autoload_broken':\n"
            "Collapsing section named 'autoload_broken':\n"
            "Failed converting argument 'class' to callable:\n"
            "'spork' is not callable",
            central.ConfigManager, [{
                    'autoload_broken': basics.HardCodedConfigSection({
                            'class': 'spork'})}])

    def test_autoload_wrong_type(self):
        self.check_error(
            "Section 'autoload_wrong' is marked as autoload but type is "
            'drawer, not configsection',
            central.ConfigManager, [{
                    'autoload_wrong': basics.HardCodedConfigSection({
                            'class': drawer})}])

    def test_lazy_refs(self):
        @configurable({'myrepo': 'lazy_ref:repo', 'thing': 'lazy_ref'},
                      typename='repo')
        def reporef(myrepo=None, thing=None):
            return myrepo, thing
        @configurable({'myrepo': 'lazy_refs:repo', 'thing': 'lazy_refs'},
                      typename='repo')
        def reporefs(myrepo=None, thing=None):
            return myrepo, thing
        @configurable(typename='repo')
        def myrepo():
            return 'repo!'
        manager = central.ConfigManager([{
                    'myrepo': basics.HardCodedConfigSection({'class': myrepo}),
                    'drawer': basics.HardCodedConfigSection({'class': drawer}),
                    'right':  basics.AutoConfigSection({'class': reporef,
                                                        'myrepo': 'myrepo'}),
                    'wrong':  basics.AutoConfigSection({'class': reporef,
                                                        'myrepo': 'drawer'}),
                    }])
        self.check_error(
            "reference 'drawer' should be of type 'repo', got 'drawer'",
            manager.objects.repo['wrong'][0].collapse)
        self.assertEqual('repo!', manager.objects.repo['right'][0].instantiate())

        manager = central.ConfigManager([{
                    'myrepo': basics.HardCodedConfigSection({'class': myrepo}),
                    'drawer': basics.HardCodedConfigSection({'class': drawer}),
                    'right':  basics.AutoConfigSection({'class': reporefs,
                                                        'myrepo': 'myrepo'}),
                    'wrong':  basics.AutoConfigSection({'class': reporefs,
                                                        'myrepo': 'drawer'}),
                    }])
        self.check_error(
            "reference 'drawer' should be of type 'repo', got 'drawer'",
            manager.objects.repo['wrong'][0][0].collapse)
        self.assertEqual(
            ['repo!'],
            [c.instantiate() for c in manager.objects.repo['right'][0]])

    def test_inherited_default(self):
        manager = central.ConfigManager([{
                    'default': basics.HardCodedConfigSection({
                            'default': True,
                            'inherit': ['basic'],
                            }),
                    'uncollapsable': basics.HardCodedConfigSection({
                            'default': True,
                            'inherit': ['spork'],
                            'inherit-only': True,
                            }),
                    'basic': basics.HardCodedConfigSection({'class': drawer}),
                    }], [RemoteSource()])
        self.assertTrue(manager.get_default('drawer'))

    def test_section_names(self):
        manager = central.ConfigManager([{
                    'thing': basics.HardCodedConfigSection({'class': drawer}),
                    }], [RemoteSource()])
        collapsed = manager.collapse_named_section('thing')
        self.assertEqual('thing', collapsed.name)

    def test_inherit_only(self):
        manager = central.ConfigManager([{
                    'source': basics.HardCodedConfigSection({
                            'class': drawer,
                            'inherit-only': True,
                            }),
                    'target': basics.HardCodedConfigSection({
                            'inherit': ['source'],
                            })}], [RemoteSource()])
        self.check_error(
            "Collapsing section named 'source':\n"
            'cannot collapse inherit-only section',
            manager.collapse_named_section, 'source')
        self.assertTrue(manager.collapse_named_section('target'))

    def test_self_inherit(self):
        section = basics.HardCodedConfigSection({'inherit': ['self']})
        manager = central.ConfigManager([{
                    'self': basics.ConfigSectionFromStringDict({
                            'class': 'tests.config.test_central.drawer',
                            'inherit': 'self'}),
                    }], [RemoteSource()])
        self.check_error(
            "Collapsing section named 'self':\n"
            "Self-inherit 'self' cannot be found",
            self.get_config_obj, manager, 'drawer', 'self')
        self.check_error(
            "Self-inherit 'self' cannot be found",
            manager.collapse_section, [section])

        manager = central.ConfigManager([{
                    'self': basics.HardCodedConfigSection({
                            'inherit': ['self'],
                            })}, {
                    'self': basics.HardCodedConfigSection({
                            'inherit': ['self'],
                            })}, {
                    'self': basics.HardCodedConfigSection({
                            'class': drawer})}])
        self.assertTrue(manager.collapse_named_section('self'))
        self.assertTrue(manager.collapse_section([section]))

    def test_prepend_inherit(self):
        manager = central.ConfigManager([{
                    'sect': basics.HardCodedConfigSection({
                            'inherit.prepend': ['self']})}])
        self.check_error(
            "Collapsing section named 'sect':\n"
            'Prepending or appending to the inherit list makes no sense',
            manager.collapse_named_section, 'sect')

    def test_list_prepend(self):
        @configurable({'seq': 'list'})
        def seq(seq):
            return seq
        manager = central.ConfigManager([{
                    'inh': basics.HardCodedConfigSection({
                            'inherit': ['sect'],
                            'seq.prepend': ['pre'],
                            }),
                    'sect': basics.HardCodedConfigSection({
                            'inherit': ['base'],
                            'seq': ['1', '2'],
                            })}, {
                    'base': basics.HardCodedConfigSection({
                            'class': seq,
                            'seq.prepend': ['-1'],
                            'seq.append': ['post'],
                            })}])
        self.assertEqual(['-1', 'post'], manager.objects.seq['base'])
        self.assertEqual(['1', '2'], manager.objects.seq['sect'])
        self.assertEqual(['pre', '1', '2'], manager.objects.seq['inh'])

    def test_str_prepend(self):
        @configurable({'string': 'str'})
        def sect(string):
            return string
        manager = central.ConfigManager([{
                    'inh': basics.HardCodedConfigSection({
                            'inherit': ['sect'],
                            'string.prepend': 'pre',
                            }),
                    'sect': basics.HardCodedConfigSection({
                            'inherit': ['base'],
                            'string': 'b',
                            })}, {
                    'base':
                        basics.HardCodedConfigSection({
                            'class': sect,
                            'string.prepend': 'a',
                            'string.append': 'c',
                            })}])
        self.assertEqual('a c', manager.objects.sect['base'])
        self.assertEqual('b', manager.objects.sect['sect'])
        self.assertEqual('pre b', manager.objects.sect['inh'])



"""DepSet parsing.

Turns a DepSet (depend, rdepend, SRC_URI, license, etc) into
appropriate conditionals.
"""

__all__ = ("DepSet", "stringify_boolean")

from snakeoil.compatibility import IGNORED_EXCEPTIONS
from snakeoil.iterables import expandable_chain
from snakeoil.sequences import iflatten_instance

from ..restrictions import boolean, packages, restriction, values
from .atom import atom, transitive_use_atom
from .errors import DepsetParseError


class DepSet(boolean.AndRestriction):
    """Gentoo DepSet syntax parser"""

    __slots__ = ('element_class', '_node_conds', '_known_conditionals')

    _evaluate_collapse = True

    # do not enable instance caching w/out adjust evaluate_depset!
    __inst_caching__ = False

    def __init__(self, restrictions='', element_class=atom,
                 node_conds=True, known_conditionals=None):
        sf = object.__setattr__
        sf(self, '_known_conditionals', known_conditionals)
        sf(self, 'element_class', element_class)
        sf(self, 'restrictions', restrictions)
        sf(self, '_node_conds', node_conds)
        sf(self, 'type', restriction.package_type)
        sf(self, 'negate', False)

    @classmethod
    def parse(cls, dep_str, element_class,
              operators=None, attr=None,
              element_func=None, transitive_use_atoms=False,
              allow_src_uri_file_renames=False):
        """
        :param dep_str: string abiding by DepSet syntax
        :param operators: mapping of node -> callable for special operators
            in DepSet syntax
        :param element_func: if None, element_class is used for generating
            elements, else it's used to generate elements.
            Mainly useful for when you need to curry a few args for instance
            generation, since element_class _must_ be a class
        :param element_class: class of generated elements
        :param attr: name of the DepSet attribute being parsed
        """
        if element_func is None:
            element_func = element_class

        restrictions = []
        if operators is None:
            operators = {"||": boolean.OrRestriction, "": boolean.AndRestriction}

        raw_conditionals = []
        depsets = [restrictions]

        node_conds = False
        words = iter(dep_str.split())
        # we specifically do it this way since expandable_chain has a bit of nasty
        # overhead to the tune of 33% slower
        if allow_src_uri_file_renames:
            words = expandable_chain(words)
        k = None
        try:
            for k in words:
                if ")" == k:
                    # no elements == error. if closures don't map up,
                    # indexerror would be chucked from trying to pop
                    # the frame so that is addressed.
                    if not depsets[-1] or not raw_conditionals:
                        raise DepsetParseError(dep_str, attr=attr)
                    elif raw_conditionals[-1] in operators:
                        if len(depsets[-1]) == 1:
                            depsets[-2].append(depsets[-1][0])
                        else:
                            depsets[-2].append(
                                operators[raw_conditionals[-1]](*depsets[-1]))
                    else:
                        node_conds = True
                        c = raw_conditionals[-1]
                        if c[0] == "!":
                            c = values.ContainmentMatch(c[1:-1], negate=True)
                        else:
                            c = values.ContainmentMatch(c[:-1])

                        depsets[-2].append(
                            packages.Conditional("use", c, tuple(depsets[-1])))

                    raw_conditionals.pop()
                    depsets.pop()

                elif "(" == k:
                    k = ''
                    # push another frame on
                    depsets.append([])
                    raw_conditionals.append(k)

                elif k[-1] == '?' or k in operators:
                    # use conditional or custom op.
                    # no tokens left == bad dep_str.
                    k2 = next(words)

                    if k2 != "(":
                        raise DepsetParseError(dep_str, k2, attr=attr)

                    # push another frame on
                    depsets.append([])
                    raw_conditionals.append(k)

                elif "|" in k:
                    raise DepsetParseError(dep_str, k, attr=attr)
                elif allow_src_uri_file_renames:
                    try:
                        k2 = next(words)
                    except StopIteration:
                        depsets[-1].append(element_func(k))
                    else:
                        if k2 != '->':
                            depsets[-1].append(element_func(k))
                            words.appendleft((k2,))
                        else:
                            k3 = next(words)
                            # file rename
                            depsets[-1].append(element_func(k, k3))
                else:
                    # node/element
                    depsets[-1].append(element_func(k))

        except IGNORED_EXCEPTIONS:
            raise
        except DepsetParseError:
            # [][-1] for a frame access, which means it was a parse error.
            raise
        except StopIteration:
            if k is None:
                raise
            raise DepsetParseError(dep_str, k, attr=attr)
        except Exception as e:
            raise DepsetParseError(dep_str, e, attr=attr) from e

        # check if any closures required
        if len(depsets) != 1:
            raise DepsetParseError(dep_str, attr=attr)

        if transitive_use_atoms and not node_conds:
            # localize to this scope for speed.
            element_class = transitive_use_atom
            # we can't rely on iter(self) here since it doesn't
            # descend through boolean restricts.
            node_conds = cls._has_transitive_use_atoms(restrictions)

        return cls(tuple(restrictions), element_class, node_conds)

    @staticmethod
    def _has_transitive_use_atoms(iterable):
        kls = transitive_use_atom
        ifunc = isinstance
        return any(ifunc(x, kls) for x in iflatten_instance(iterable, atom))

    def evaluate_depset(self, cond_dict, tristate_filter=None, pkg=None):
        """
        :param cond_dict: container to be used for conditional collapsing,
            typically is a use list
        :param tristate_filter: a control; if specified, must be a container
            of conditionals to lock to cond_dict.
            during processing, if it's not in tristate_filter will
            automatically enable the payload
            (regardless of the conditionals negation)
        """
        if not self.has_conditionals:
            return self

        results = []
        self.evaluate_conditionals(
            self.__class__, results,
            cond_dict, tristate_filter, force_collapse=True)

        return self.__class__(tuple(results), self.element_class, False)

    @staticmethod
    def find_cond_nodes(restriction_set, yield_non_conditionals=False):
        conditions_stack = []
        new_set = expandable_chain(restriction_set)
        for cur_node in new_set:
            if isinstance(cur_node, packages.Conditional):
                conditions_stack.append(cur_node.restriction)
                new_set.appendleft(list(cur_node.payload) + [None])
            elif isinstance(cur_node, transitive_use_atom):
                new_set.appendleft(cur_node.convert_to_conditionals())
            elif (isinstance(cur_node, boolean.base) and
                    not isinstance(cur_node, atom)):
                new_set.appendleft(cur_node.restrictions)
            elif cur_node is None:
                conditions_stack.pop()
            elif conditions_stack or yield_non_conditionals: # leaf
                yield (cur_node, conditions_stack[:])

    @property
    def node_conds(self):
        if self._node_conds is False:
            object.__setattr__(self, "_node_conds", {})
        elif self._node_conds is True:
            nc = {}

            always_required = set()

            for payload, restrictions in self.find_cond_nodes(self.restrictions, True):
                if not restrictions:
                    always_required.add(payload)
                else:
                    if len(restrictions) == 1:
                        current = restrictions[0]
                    else:
                        current = values.AndRestriction(*restrictions)

                    nc.setdefault(payload, []).append(current)

            for k in always_required:
                if k in nc:
                    del nc[k]
            for k in nc:
                nc[k] = tuple(nc[k])

            object.__setattr__(self, "_node_conds", nc)

        return self._node_conds

    @property
    def has_conditionals(self):
        return bool(self._node_conds)

    @property
    def known_conditionals(self):
        if self._node_conds is False:
            return frozenset()
        if self._known_conditionals is None:
            kc = set()
            for payload, restrictions in self.find_cond_nodes(self.restrictions):
                kc.update(iflatten_instance(x.vals for x in restrictions))
            kc = frozenset(kc)
            object.__setattr__(self, "_known_conditionals", kc)
            return kc
        return self._known_conditionals

    def match(self, *a):
        raise NotImplementedError

    def slotdep_str(self, domain):
        return stringify_boolean(self, domain=domain)

    force_False = force_True = match

    def __str__(self):
        return stringify_boolean(self)

    # parent __hash__() isn't inherited when __eq__() is defined in the child class
    __hash__ = boolean.AndRestriction.__hash__

    def __eq__(self, other):
        if isinstance(other, DepSet):
            return set(self.restrictions) == set(other.restrictions)
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __iter__(self):
        return iter(self.restrictions)

    def __getitem__(self, key):
        return self.restrictions[key]


def stringify_boolean(node, func=str, domain=None):
    """func is used to stringify the actual content. Useful for fetchables."""
    l = []
    if isinstance(node, DepSet):
        for x in node.restrictions:
            _internal_stringify_boolean(x, domain, func, l.append)
    else:
        _internal_stringify_boolean(node, domain, func, l.append)
    return ' '.join(l)

def _internal_stringify_boolean(node, domain, func, visit):
    """func is used to stringify the actual content. Useful for fetchables."""

    if isinstance(node, boolean.OrRestriction):
        visit("|| (")
        iterable = node.restrictions
    elif (isinstance(node, boolean.AndRestriction) and
            not isinstance(node, atom)):
        visit("(")
        iterable = node.restrictions
    elif isinstance(node, packages.Conditional):
        assert len(node.restriction.vals) == 1
        iterable = node.payload
        visit("%s%s? (" % (
            node.restriction.negate and "!" or "",
            list(node.restriction.vals)[0]))
    else:
        if (domain is not None and
                (isinstance(node, atom) and node.slot_operator == '=')):
            pkg = max(sorted(domain.all_installed_repos.itermatch(node)))
            object.__setattr__(node, "slot", pkg.slot)
            object.__setattr__(node, "subslot", pkg.subslot)
        visit(func(node))
        return
    for node in iterable:
        _internal_stringify_boolean(node, domain, func, visit)
    visit(")")



from collections import OrderedDict
from functools import partial

from pkgcore.repository.multiplex import tree
from pkgcore.repository.util import SimpleTree
from pkgcore.restrictions import packages, values

rev_sorted = partial(sorted, reverse=True)


class TestMultiplex:

    kls = staticmethod(tree)
    tree1_pkgs = (
        ("dev-util/diffball", ["1.0", "0.7"]),
        ("dev-lib/fake", ["1.0", "1.0-r1"]))
    tree2_pkgs = (
        ("dev-util/diffball", ["1.0", "1.1"]),
        ("dev-lib/bsdiff", ["1.0", "2.0"]))
    tree1_list = [f"{k}-{ver}" for k, v in tree1_pkgs for ver in v]
    tree2_list = [f"{k}-{ver}" for k, v in tree2_pkgs for ver in v]

    def setup_method(self):
        self.d1, self.d2 = {}, {}
        for key, ver in self.tree1_pkgs:
            cat, pkg = key.rsplit("/", 1)
            self.d1.setdefault(cat, {}).setdefault(pkg, []).extend(ver)
        for key, ver in self.tree2_pkgs:
            cat, pkg = key.rsplit("/", 1)
            self.d2.setdefault(cat, {}).setdefault(pkg, []).extend(ver)

        self.d1 = OrderedDict(
            (k, OrderedDict(self.d1[k].items()))
            for k in sorted(self.d1, reverse=True))
        self.d2 = OrderedDict(
            (k, OrderedDict(self.d2[k].items()))
            for k in sorted(self.d2, reverse=True))
        self.tree1 = SimpleTree(self.d1)
        self.tree2 = SimpleTree(self.d2)
        self.ctree = self.kls(self.tree1, self.tree2)

    def test_iter(self):
        assert sorted(x.cpvstr for x in self.ctree) == \
            sorted(self.tree1_list + self.tree2_list)

    def test_itermatch(self):
        imatch = self.ctree.itermatch
        assert sorted(x.cpvstr for x in imatch(packages.AlwaysTrue)) == \
            sorted(self.tree1_list + self.tree2_list)
        p = packages.PackageRestriction("package", values.StrExactMatch("diffball"))
        assert sorted(x.cpvstr for x in imatch(p)) == \
            [y for y in sorted(self.tree1_list + self.tree2_list) if "/diffball" in y]

    def test_sorting(self):
        assert list(x.cpvstr for x in self.ctree.itermatch(packages.AlwaysTrue, sorter=rev_sorted)) == \
            rev_sorted(self.tree1_list + self.tree2_list)



"""
ebuild tree manifest/digest support
"""

__all__ = ("parse_manifest", "Manifest")

import errno
import operator
import os

from snakeoil.chksum import get_handler
from snakeoil.mappings import ImmutableDict

from .. import gpg
from ..fs.livefs import iter_scan
from ..package import errors
from . import cpv


def _write_manifest(handle, chf, filename, chksums):
    """Convenient, internal method for writing manifests"""
    size = chksums.pop("size")
    handle.write("%s %s %i" % (chf.upper(), filename, size))
    for chf in sorted(chksums):
        handle.write(" %s %s" % (chf.upper(), get_handler(chf).long2str(chksums[chf])))
    handle.write('\n')


def convert_chksums(iterable):
    for chf, sum in iterable:
        chf = chf.lower()
        if chf == 'size':
            # explicit size entries are stupid, format has implicit size
            continue
        else:
            yield chf, int(sum, 16)


def parse_manifest(source, ignore_gpg=True):
    types = {"DIST": {}, "AUX": {}, "EBUILD": {}, "MISC": {}}
    # manifest v2 format: (see glep 44 for exact rules)
    # TYPE filename size (CHF sum)+
    # example 'type' entry, all one line
    # MISC metadata.xml 219 RMD160 613195ece366b33606e71ff1753be048f2507841 SHA1 d162fb909241ef50b95a3539bdfcde95429bdf81 SHA256 cbd3a20e5c89a48a842f7132fe705bf39959f02c1025052efce8aad8a8baa8dc
    # manifest v1 format is
    # CHF sum filename size
    # note that we do _not_ support manifest1
    chf_types = set(["size"])
    f = None
    try:
        if isinstance(source, str):
            i = f = open(source, "r", 32768)
        else:
            i = f = source.text_fileobj()
        if ignore_gpg:
            i = gpg.skip_signatures(f)
        for data in i:
            line = data.split()
            if not line:
                continue
            d = types.get(line[0])
            if d is None:
                raise errors.ParseChksumError(
                    source, f"unknown manifest type: {line[0]}: {line!r}")
            if len(line) % 2 != 1:
                raise errors.ParseChksumError(
                    source,
                    "manifest 2 entry doesn't have right "
                    "number of tokens, %i: %r" %
                    (len(line), line))
            chf_types.update(line[3::2])
            # this is a trick to do pairwise collapsing;
            # [size, 1] becomes [(size, 1)]
            i = iter(line[3:])
            d[line[1]] = [("size", int(line[2]))] + list(convert_chksums(zip(i, i)))
    except (IndexError, ValueError):
        raise errors.ParseChksumError(source, 'invalid data format')
    finally:
        if f is not None and f.close:
            f.close()

    for t, d in types.items():
        types[t] = ImmutableDict((k, dict(v)) for k, v in d.items())
    # ordering annoyingly matters. bad api.
    return [types[x] for x in ("DIST", "AUX", "EBUILD", "MISC")]


class Manifest:

    def __init__(self, path, enforce_gpg=False, thin=False, allow_missing=False):
        self.path = path
        self.thin = thin
        self.allow_missing = allow_missing
        self._gpg = enforce_gpg
        self._sourced = False

    def _pull_manifest(self):
        if self._sourced:
            return
        try:
            data = parse_manifest(self.path, ignore_gpg=self._gpg)
        except EnvironmentError as e:
            if not (self.thin or self.allow_missing) or e.errno != errno.ENOENT:
                raise errors.ParseChksumError(self.path, e) from e
            data = {}, {}, {}, {}
        except errors.ChksumError as e:
            # recreate cpv from manifest path
            catpn = os.sep.join(self.path.split(os.sep)[-3:-1])
            pkg = cpv.UnversionedCPV(catpn)
            raise errors.MetadataException(pkg, 'manifest', str(e))
        self._dist, self._aux, self._ebuild, self._misc = data
        self._sourced = True

    def update(self, fetchables, chfs=None):
        """Update the related Manifest file.

        :param fetchables: fetchables of the package
        """

        if self.thin and not fetchables:
            # Manifest files aren't necessary with thin manifests and no distfiles
            return

        _key_sort = operator.itemgetter(0)

        excludes = frozenset(["CVS", ".svn", "Manifest"])
        aux, ebuild, misc = {}, {}, {}
        if not self.thin:
            filesdir = '/files/'
            for obj in iter_scan('/', offset=os.path.dirname(self.path), chksum_types=chfs):
                if not obj.is_reg:
                    continue
                pathname = obj.location
                if excludes.intersection(pathname.split('/')):
                    continue
                if pathname.startswith(filesdir):
                    d = aux
                    pathname = pathname[len(filesdir):]
                elif obj.dirname == '/':
                    pathname = pathname[1:]
                    if obj.location[-7:] == '.ebuild':
                        d = ebuild
                    else:
                        d = misc
                else:
                    raise Exception("Unexpected directory found in %r; %r" % (self.path, obj.dirname))
                d[pathname] = dict(obj.chksums)

        handle = open(self.path, 'w')

        # write it in alphabetical order; aux gets flushed now.
        for path, chksums in sorted(aux.items(), key=_key_sort):
            _write_manifest(handle, 'AUX', path, chksums)

        # next dist...
        for fetchable in sorted(fetchables, key=operator.attrgetter('filename')):
            _write_manifest(
                handle, 'DIST', os.path.basename(fetchable.filename),
                dict(fetchable.chksums))

        # then ebuild and misc
        for mtype, inst in (("EBUILD", ebuild), ("MISC", misc)):
            for path, chksum in sorted(inst.items(), key=_key_sort):
                _write_manifest(handle, mtype, path, chksum)

    @property
    def aux_files(self):
        self._pull_manifest()
        return self._aux

    @property
    def distfiles(self):
        self._pull_manifest()
        return self._dist

    @property
    def ebuilds(self):
        self._pull_manifest()
        return self._ebuild

    @property
    def misc(self):
        self._pull_manifest()
        return self._misc



import pytest
from pkgcore.ebuild.atom import atom
from pkgcore.repository.util import SimpleTree
from pkgcore.scripts import pmerge
from pkgcore.test.misc import FakePkg, FakeRepo
from pkgcore.util.parserestrict import parse_match


# TODO: make repo objs into configurable fixtures
class TestTargetParsing:

    def test_base_targets(self):
        repo = SimpleTree({'spork': {'foon': ('1', '1.0.1', '2')}})
        installed_repos = SimpleTree({'foo': {'bar': ('1',)}})
        for cat in ('', 'spork/'):
            a = pmerge.parse_target(parse_match(f'={cat}foon-1'), repo, installed_repos)
            assert len(a) == 1
            assert a[0].key == 'spork/foon'
            assert [x.fullver for x in repo.itermatch(a[0])] == ['1']
            a = pmerge.parse_target(parse_match(f'{cat}foon'), repo, installed_repos)
            assert len(a) == 1
            assert a[0].key == 'spork/foon'
            assert (
                sorted(x.fullver for x in repo.itermatch(a[0])) ==
                sorted(['1', '1.0.1', '2'])
            )

    def test_no_matches(self):
        repo = SimpleTree({
            'spork': {'foon': ('1',)},
            'spork2': {'foon': ('2',)}})
        installed_repos = SimpleTree({'foo': {'bar': ('1',)}})
        with pytest.raises(pmerge.NoMatches):
            pmerge.parse_target(parse_match("foo"), repo, installed_repos)

    def test_ambiguous(self):
        repo = SimpleTree({
            'spork': {'foon': ('1',)},
            'spork2': {'foon': ('2',)}})
        installed_repos = SimpleTree({'foo': {'bar': ('1',)}})
        with pytest.raises(pmerge.AmbiguousQuery):
            pmerge.parse_target(parse_match("foon"), repo, installed_repos)

    def test_globbing(self):
        repo = SimpleTree({
            'spork': {'foon': ('1',)},
            'spork2': {'foon': ('2',)}})
        installed_repos = SimpleTree({'foo': {'bar': ('1',)}})
        a = pmerge.parse_target(parse_match('*/foon'), repo, installed_repos)
        assert len(a) == 2

    def test_collision_repo(self):
        # test pkg name collision between real and virtual pkgs in a repo, but not installed
        # repos, the real pkg will be selected over the virtual
        installed_repos = SimpleTree({'foo': {'baz': ('1',)}})
        repo = SimpleTree({'foo': {'bar': ('1',)}, 'virtual': {'bar': ('1',)}})
        a = pmerge.parse_target(parse_match("bar"), repo, installed_repos)
        assert len(a) == 1
        assert a[0].key == 'foo/bar'
        assert [x.key for x in repo.match(a[0])] == ['foo/bar']

    def test_collision_livefs(self):
        # test pkg name collision between real and virtual pkgs on livefs
        # repos, the real pkg will be selected over the virtual
        installed_repos = SimpleTree({'foo': {'bar': ('1',)}, 'virtual': {'bar': ('0',)}})
        repo = SimpleTree({'foo': {'bar': ('1',)}, 'virtual': {'bar': ('1',)}})
        a = pmerge.parse_target(parse_match("bar"), repo, installed_repos)
        assert len(a) == 1
        assert a[0].key == 'foo/bar'
        assert [x.key for x in repo.match(a[0])] == ['foo/bar']

    def test_collision_slotted(self):
        pkgs = [
            FakePkg('foo/bar-1.0.1', slot='0'),
            FakePkg('foo/bar-2.0.2', slot='2'),
            FakePkg('foon/bar-3.4.5', slot='0'),
        ]
        installed_pkgs = [
            FakePkg('foo/bar-1.0.0', slot='0'),
            FakePkg('foo/bar-2.0.1', slot='2'),
        ]
        installed_repos = FakeRepo(pkgs=installed_pkgs)
        repo = FakeRepo(pkgs=pkgs)
        a = pmerge.parse_target(parse_match("bar:0"), repo, installed_repos)
        assert len(a) == 1
        assert a[0].key == 'foo/bar'
        assert a[0].match(atom('foo/bar:0'))
        assert not a[0].match(atom('foo/bar:2'))



from pkgcore.cache import errors
from snakeoil.chksum import LazilyHashedPath

generic_data = \
    ("sys-libs/libtrash-2.4",
        (('DEPEND', 'virtual/libc dev-lang/perl'),
        ('DESCRIPTION', 'provides a trash can by intercepting calls...'),
        ('EAPI', ''),
        ('HOMEPAGE', 'http://pages.stern.nyu.edu/~marriaga/software/libtrash/'),
        ('IUSE', ''),
        ('KEYWORDS', '~amd64 ~ppc ~x86'),
        ('LICENSE', 'GPL-2'),
        ('PDEPEND', ''),
        ('RDEPEND', 'virtual/libc dev-lang/perl'),
        ('RESTRICT', ''),
        ('SLOT', '0'),
        ('SRC_URI', 'http://pages.stern.nyu.edu/~marriaga/software/blah.tgz'),
        ('_eclasses_',
            {
                'toolchain-funcs': LazilyHashedPath('/var/gentoo/repos/gentoo/eclass', mtime=1155996352),
                'multilib': LazilyHashedPath('/var/gentoo/repos/gentoo/eclass', mtime=1156014349),
                'eutils': LazilyHashedPath('/var/gentoo/repos/gentoo/eclass', mtime=1155996352),
                'portability': LazilyHashedPath('/var/gentoo/repos/gentoo/eclass', mtime=1141850196)
            }
        ),
        ('_mtime_', 1000),
    ),
)

class GenericCacheMixin:

    cache_keys = ("DEPENDS", "RDEPEND", "EAPI", "HOMEPAGE", "KEYWORDS",
        "LICENSE", "PDEPEND", "RESTRICT", "SLOT", "SRC_URI",
        "_eclasses_", "_mtime_")

    # truncating the original metadata we grabbed for 80 char...
    test_data = (generic_data,)

    def get_db(self, readonly=False):
        raise NotImplementedError(
            self, "get_db- must be overridden for test mixin, "
            "setting self.db to a cache instance ")

    def test_readonly(self):
        db = self.get_db(False)
        for key, raw_data in self.test_data:
            d = dict(raw_data)
            db[key] = d
        db.commit()
        db = self.get_db(True)
        for key, raw_data in self.test_data:
            d = dict(raw_data)
            self.assertRaises(errors.ReadOnly, db.__setitem__, key, d)

    def test_setitem(self):
        db = self.get_db(False)
        for key, raw_data in self.test_data:
            d = dict(raw_data)
            db[key] = d



"""
cache subsystem exceptions
"""

__all__ = (
    "CacheError", "InitializationError", "CacheCorruption",
    "GeneralCacheCorruption", "ReadOnly",
)

from ..exceptions import PkgcoreException


class CacheError(PkgcoreException):
    pass


class InitializationError(CacheError):
    def __init__(self, class_name, error):
        super().__init__(
            f'creation of instance {class_name} failed due to {error}')
        self.error, self.class_name = error, class_name


class CacheCorruption(CacheError):
    def __init__(self, key, ex):
        super().__init__(f'{key} is corrupt: {ex}')
        self.key, self.ex = key, ex


class GeneralCacheCorruption(CacheError):
    def __init__(self, ex):
        super().__init__(f'corruption detected: {ex}')
        self.ex = ex


class ReadOnly(CacheError):
    def __init__(self, info=''):
        super().__init__(f'cache is non-modifiable {info}')
        self.info = info



"""repository inspection interface

pinspect is used to extract various information from repos. For example,
it can perform aggregated EAPI, license, eclass, and mirror usage queries
across specified repos. Any repo type can be queried, e.g. ebuild,
binary, or vdb.

It also provides an interface to all profile specific metadata, e.g. package
masks, inheritance trees, etc. This makes it easier to inspect profile
differences without sorting through the inheritance tree and reading the raw
files.

Finally, a portageq compatible interface is provided for several commands that
were historically used in ebuilds.
"""

from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from itertools import groupby, islice
from operator import attrgetter, itemgetter

from snakeoil.cli import arghparse
from snakeoil.sequences import iflatten_instance, unstable_unique

from .. import fetch
from ..ebuild import inspect_profile
from ..ebuild import portageq as _portageq
from ..package import errors
from ..restrictions import packages
from ..util import commandline

pkgcore_opts = commandline.ArgumentParser(domain=False, script=(__file__, __name__))
argparser = commandline.ArgumentParser(
    suppress=True, description=__doc__, parents=(pkgcore_opts,))
subparsers = argparser.add_subparsers(description="report applets")

pkgsets = subparsers.add_parser(
    "pkgsets", description="pkgset related introspection")
mux = pkgsets.add_mutually_exclusive_group()
mux.add_argument(
    "--all", action='store_true', default=False,
    help="display info on all pkgsets")
mux.add_argument(
    "pkgsets", nargs="*", metavar="pkgset", default=[],
    action=commandline.StoreConfigObject,
    config_type='pkgset', store_name=True,
    help="pkgset to inspect")
del mux
@pkgsets.bind_main_func
def pkgsets_run(opts, out, err):
    if not opts.pkgsets:
        if not opts.all:
            out.write(out.bold, 'available pkgset(s): ', out.reset,
                      ', '.join(repr(x) for x in sorted(opts.config.pkgset)))
            return 0
        else:
            opts.pkgsets = sorted(opts.config.pkgset)

    for position, (set_name, pkgset) in enumerate(opts.pkgsets):
        if position:
            out.write()
        out.write(out.bold, 'pkgset ', repr(set_name), out.reset, ':')
        out.first_prefix.append('  ')
        for restrict in sorted(pkgset):
            out.write(restrict)
        out.first_prefix.pop()
    return 0


def print_simple_histogram(data, out, format, total, sort_by_key=False,
                           first=None, last=None):
    # do the division up front...
    total = float(total) / 100

    if sort_by_key:
        data = sorted(data.items(), key=itemgetter(0))
    else:
        data = sorted(data.items(), key=itemgetter(1), reverse=True)

    if first:
        data = islice(data, 0, first)
    elif last:
        data = list(data)[-last:]

    for key, val in data:
        out.write(format %
                  {'key': str(key), 'val': val,
                   'percent': "%2.2f%%" % (val/total,)})


class histo_data(arghparse.ArgparseCommand):

    per_repo_summary = None
    allow_no_detail = False

    def bind_to_parser(self, parser):
        mux = parser.add_mutually_exclusive_group()
        mux.add_argument(
            "--no-final-summary", action='store_true', default=False,
            help="disable outputting a summary of data across all repos")

        parser.set_defaults(repo_summary=bool(self.per_repo_summary))
        if self.per_repo_summary:
            mux.add_argument(
                "--no-repo-summary", dest='repo_summary',
                action='store_false',
                help="disable outputting repo summaries")

        parser.set_defaults(no_detail=False)
        if self.allow_no_detail:
            mux.add_argument(
                "--no-detail", action='store_true', default=False,
                help="disable outputting a detail view of all repos")

        parser.add_argument(
            "--sort-by-name", action='store_true', default=False,
            help="sort output by name, rather then by frequency")

        mux = parser.add_mutually_exclusive_group()

        mux.add_argument(
            "--first", action="store", type=int, default=0,
            help="show only the first N detail items")

        mux.add_argument(
            "--last", action="store", type=int, default=0,
            help="show only the last N detail items")

        parser.add_argument(
            "repos", metavar='repo', nargs='*',
            action=commandline.StoreRepoObject, store_name=True,
            default=commandline.CONFIG_ALL_DEFAULT,
            help="repo(s) to inspect")

        arghparse.ArgparseCommand.bind_to_parser(self, parser)

    def get_data(self, repo, options):
        raise NotImplementedError()

    def transform_data_to_detail(self, data):
        return data

    def transform_data_to_summary(self, data):
        return data

    def __call__(self, opts, out, err):
        global_stats = {}
        position = 0
        total_pkgs = 0
        for repo_name, repo in opts.repos:
            if position:
                out.write()
            position += 1
            out.write(out.bold, "repository", out.reset, ' ',
                      repr(repo_name), ':')
            data, repo_total = self.get_data(repo, opts)
            detail_data = self.transform_data_to_detail(data)
            if not opts.no_detail:
                out.first_prefix.append("  ")
                if not data:
                    out.write("no pkgs found")
                else:
                    print_simple_histogram(
                        detail_data, out, self.per_repo_format,
                        repo_total, sort_by_key=opts.sort_by_name,
                        first=opts.first, last=opts.last)
                out.first_prefix.pop()
            for key, val in detail_data.items():
                global_stats.setdefault(key, 0)
                global_stats[key] += val
            total_pkgs += repo_total

            if not opts.repo_summary:
                continue
            out.write(
                out.bold, 'summary', out.reset, ': ',
                self.per_repo_summary %
                self.transform_data_to_summary(data))

        if position > 1 and not opts.no_final_summary:
            out.write()
            out.write(out.bold, 'summary', out.reset, ':')
            out.first_prefix.append('  ')
            print_simple_histogram(
                global_stats, out, self.summary_format,
                total_pkgs, sort_by_key=opts.sort_by_name)
            out.first_prefix.pop()
        return 0


class eapi_usage_kls(histo_data):

    per_repo_format = ("eapi: %(key)r %(val)s pkgs found, %(percent)s of the repo")

    summary_format = ("eapi: %(key)r %(val)s pkgs found, %(percent)s of all repos")

    def get_data(self, repo, options):
        eapis = {}
        pos = 0
        for pos, pkg in enumerate(repo):
            eapis.setdefault(str(pkg.eapi), 0)
            eapis[str(pkg.eapi)] += 1
        return eapis, pos + 1

eapi_usage = subparsers.add_parser(
    "eapi_usage", description="report of eapi usage for targeted repos")
eapi_usage.bind_class(eapi_usage_kls())


class license_usage_kls(histo_data):

    per_repo_format = "license: %(key)r %(val)s pkgs found, %(percent)s of the repo"

    summary_format = "license: %(key)r %(val)s pkgs found, %(percent)s of all repos"

    def get_data(self, repo, options):
        data = {}
        pos = 0
        for pos, pkg in enumerate(repo):
            for license in unstable_unique(iflatten_instance(pkg.license)):
                data.setdefault(license, 0)
                data[license] += 1
        return data, pos + 1

license_usage = subparsers.add_parser(
    "license_usage", description="report of license usage for targeted repos")
license_usage.bind_class(license_usage_kls())


class eclass_usage_kls(histo_data):

    per_repo_format = "eclass: %(key)r %(val)s pkgs found, %(percent)s of the repo"

    summary_format = "eclass: %(key)r %(val)s pkgs found, %(percent)s of all repos"

    def get_data(self, repo, options):
        pos, data = 0, defaultdict(lambda:0)
        for pos, pkg in enumerate(repo):
            for eclass in getattr(pkg, 'inherited', ()):
                data[eclass] += 1
        return data, pos + 1

eclass_usage = subparsers.add_parser(
    "eclass_usage", description="report of eclass usage for targeted repos")
eclass_usage.bind_class(eclass_usage_kls())


class mirror_usage_kls(histo_data):

    per_repo_format = "mirror: %(key)r %(val)s pkgs found, %(percent)s of the repo"

    summary_format = "mirror: %(key)r %(val)s pkgs found, %(percent)s of all repos"

    def get_data(self, repo, options):
        data = {}
        for pos, pkg in enumerate(repo):
            for fetchable in iflatten_instance(pkg.fetchables, fetch.fetchable):
                for mirror in fetchable.uri.visit_mirrors(treat_default_as_mirror=False):
                    if isinstance(mirror, tuple):
                        mirror = mirror[0]
                    data.setdefault(mirror.mirror_name, 0)
                    data[mirror.mirror_name] += 1
        return data, pos + 1

mirror_usage = subparsers.add_parser(
    "mirror_usage", description="report of SRC_URI mirror usage for targeted repos")
mirror_usage.bind_class(mirror_usage_kls())


class distfiles_usage_kls(histo_data):

    per_repo_format = "package: %(key)r %(val)s bytes, referencing %(percent)s of the unique total"

    per_repo_summary = "unique total %(total)i bytes, sharing %(shared)i bytes"

    summary_format = "package: %(key)r %(val)s pkgs found, %(percent)s of all repos"

    allow_no_detail = True

    def bind_to_parser(self, parser):
        histo_data.bind_to_parser(self, parser)
        parser.add_argument(
            "--include-nonmirrored", action='store_true', default=False,
            help="if set, nonmirrored  distfiles will be included in the total")
        parser.add_argument(
            "--include-restricted", action='store_true', default=False,
            help="if set, fetch restricted distfiles will be included in the total")

    def get_data(self, repo, options):
        owners = defaultdict(set)
        iterable = repo.itermatch(packages.AlwaysTrue, sorter=sorted)
        items = {}
        for key, subiter in groupby(iterable, attrgetter("key")):
            for pkg in subiter:
                if not options.include_restricted and 'fetch' in pkg.restrict:
                    continue
                if not options.include_nonmirrored and 'mirror' in pkg.restrict:
                    continue
                for fetchable in iflatten_instance(pkg.fetchables, fetch.fetchable):
                    owners[fetchable.filename].add(key)
                    items[fetchable.filename] = fetchable.chksums.get("size", 0)

        data = defaultdict(lambda: 0)
        for filename, keys in owners.items():
            for key in keys:
                data[key] += items[filename]
        unique = sum(items.values())
        shared = sum(items[k] for (k, v) in owners.items() if len(v) > 1)
        return (data, {"total": unique, "shared": shared}), unique

    def transform_data_to_detail(self, data):
        return data[0]

    def transform_data_to_summary(self, data):
        return data[1]


distfiles_usage = subparsers.add_parser(
    "distfiles_usage",
    description="report detailing distfiles space usage for targeted repos")
distfiles_usage.bind_class(distfiles_usage_kls())

query = subparsers.add_parser(
    "query",
    description="auxiliary access to ebuild/repo info via portageq akin api")
_portageq.bind_parser(query, name='query')

portageq = subparsers.add_parser(
    "portageq", description="portageq compatible interface to query commands")
_portageq.bind_parser(portageq, compat=True)

profile = subparsers.add_parser(
    "profile", description="profile related querying")
# TODO: restrict to ebuild repos
profile_opts = profile.add_argument_group('subcommand options')
profile_opts.add_argument(
    '-r', '--repo', metavar='REPO', help='target repo',
    action=commandline.StoreRepoObject, repo_type='config')
inspect_profile.bind_parser(profile, 'profile')


def _bad_digest(pkg):
    """Check if a given package has a broken or missing digest."""
    try:
        pkg.fetchables
    except errors.MetadataException:
        return pkg, True
    return pkg, False


digests = subparsers.add_parser(
    "digests", domain=True, description="identify what packages are missing digest info")
digests.add_argument(
    'repos', nargs='*', help="repo to inspect",
    action=commandline.StoreRepoObject, allow_external_repos=True, store_name=True)
@digests.bind_main_func
def digest_manifest(options, out, err):
    for name, repo in options.repos:
        count = 0
        broken = []
        out.write(f"inspecting {name!r} repo:")
        out.flush()

        # TODO: move to ProcessPoolExecutor once underlying pkg wrapper classes can be pickled
        with ThreadPoolExecutor() as executor:
            for pkg, bad in executor.map(_bad_digest, iter(repo)):
                count += 1
                if bad:
                    broken.append(pkg)

        if count:
            if broken:
                out.write('Packages with broken digests:')
                out.first_prefix.append("  ")
                out.later_prefix.append("  ")
                for pkg in sorted(broken):
                    out.write(pkg.cpvstr)
                out.first_prefix.pop()
                out.later_prefix.pop()
                percent = len(broken) / count * 100
                out.write(
                    f"{len(broken)} out of {count} ({round(percent, 2)}%) packages "
                    "in the repo have bad checksum data"
                )
            else:
                out.write('repo has no broken digests')
        else:
            out.write("repo has no packages")

        out.write()



"""
base repository template
"""

__all__ = (
    "CategoryIterValLazyDict", "PackageMapping", "VersionMapping", "tree"
)

import os

from snakeoil.klass import jit_attr
from snakeoil.mappings import DictMixin, LazyValDict
from snakeoil.osutils import pjoin
from snakeoil.sequences import iflatten_instance

from ..ebuild.atom import atom
from ..operations import repo
from ..restrictions import boolean, packages, restriction, values
from ..restrictions.util import collect_package_restrictions


class IterValLazyDict(LazyValDict):

    __slots__ = ()

    def __str__(self):
        return str(list(self))

    def force_regen(self, key):
        if key in self._vals:
            del self._vals[key]
        else:
            self._keys = tuple(x for x in self._keys if x != key)


class CategoryIterValLazyDict(IterValLazyDict):

    __slots__ = ()

    def force_add(self, key):
        if key not in self:
            s = set(self._keys)
            s.add(key)
            self._keys = tuple(s)

    def force_remove(self, key):
        if key in self:
            self._keys = tuple(x for x in self._keys if x != key)

    __iter__ = IterValLazyDict.keys

    def __contains__(self, key):
        if self._keys_func is not None:
            return key in list(self.keys())
        return key in self._keys


class PackageMapping(DictMixin):

    def __init__(self, parent_mapping, pull_vals):
        self._cache = {}
        self._parent = parent_mapping
        self._pull_vals = pull_vals

    def __getitem__(self, key):
        o = self._cache.get(key)
        if o is not None:
            return o
        if key not in self._parent:
            raise KeyError(key)
        self._cache[key] = vals = self._pull_vals(key)
        return vals

    def keys(self):
        return self._parent.keys()

    def __contains__(self, key):
        return key in self._cache or key in self._parent

    def force_regen(self, cat):
        try:
            del self._cache[cat]
        except KeyError:
            pass


class VersionMapping(DictMixin):

    def __init__(self, parent_mapping, pull_vals):
        self._cache = {}
        self._parent = parent_mapping
        self._pull_vals = pull_vals

    def __getitem__(self, key):
        o = self._cache.get(key)
        if o is not None:
            return o
        if not key[1] in self._parent.get(key[0], ()):
            raise KeyError(key)
        val = self._pull_vals(key)
        self._cache[key] = val
        return val

    def keys(self):
        for cat, pkgs in self._parent.items():
            for pkg in pkgs:
                yield (cat, pkg)

    def force_regen(self, key, val):
        if val:
            self._cache[key] = val
        else:
            self._cache.pop(key, None)


class tree:
    """Template for all repository variants.

    Args:
        frozen (bool): controls whether the repository is mutable or immutable

    Attributes:
        raw_repo: if wrapping a repo, set raw_repo per instance to it
        livefs (bool): set it to True if it's a repository representing a livefs
        package_class: callable to generate a package instance, must override
        configured (bool): if a repo is unusable for merging/unmerging
            without being configured, set it to False
        frozen_settable (bool): controls whether frozen is able to be set
            on initialization
        operations_kls: callable to generate a repo operations instance

        categories (dict): available categories in the repo
        packages (dict): mapping of packages to categories in the repo
        versions (dict): mapping of versions to packages in the repo
        frozen (bool): repository mutability status
        lock: TODO
    """

    raw_repo = None
    is_supported = True
    livefs = False
    package_class = None
    configured = True
    frozen_settable = True
    operations_kls = repo.operations
    pkg_masks = frozenset()

    def __init__(self, frozen=False):
        self.categories = CategoryIterValLazyDict(
            self._get_categories, self._get_categories)
        self.packages = PackageMapping(self.categories, self._get_packages)
        self.versions = VersionMapping(self.packages, self._get_versions)

        if self.frozen_settable:
            self.frozen = frozen
        self.lock = None

    def configure(self, *args):
        """Return a configured form of the repository."""
        raise NotImplementedError(self, "configure")

    def _get_categories(self, *args):
        """this must return a list, or sequence"""
        raise NotImplementedError(self, "_get_categories")

    def _get_packages(self, category):
        """this must return a list, or sequence"""
        raise NotImplementedError(self, "_get_packages")

    def _get_versions(self, package):
        """this must return a list, or sequence"""
        raise NotImplementedError(self, "_get_versions")

    def __getitem__(self, cpv):
        cpv_inst = self.package_class(*cpv)
        if cpv_inst.fullver not in self.versions[(cpv_inst.category, cpv_inst.package)]:
            raise KeyError(cpv)
        return cpv_inst

    def __setitem__(self, *vals):
        raise AttributeError

    def __delitem__(self, cpv):
        raise AttributeError

    def __iter__(self):
        """Filtered iterator over all the repo's packages.

        All packages with metadata issues are skipped.""
        """
        return self.itermatch(packages.AlwaysTrue)

    def __len__(self):
        return sum(len(v) for v in self.versions.values())

    def __contains__(self, obj):
        """Determine if a path or a package is in a repo."""
        if isinstance(obj, str):
            path = os.path.normpath(obj)
            try:
                repo_path = os.path.realpath(getattr(self, 'location'))
            except AttributeError:
                return False

            # existing relative path
            if not path.startswith(os.sep) and os.path.exists(pjoin(repo_path, path)):
                return True

            # existing full path
            fullpath = os.path.realpath(os.path.abspath(path))
            if fullpath.startswith(repo_path) and os.path.exists(fullpath):
                return True

            return False
        else:
            for pkg in self.itermatch(obj):
                return True
            return False

    def has_match(self, atom, **kwds):

        kwds.pop("sorter", None)
        kwds.pop("yield_none", None)

        for pkg in self.itermatch(atom, **kwds):
            return True
        return False

    def match(self, atom, **kwds):
        return list(self.itermatch(atom, **kwds))

    def itermatch(self, restrict, sorter=None, pkg_filter=None, versioned=True,
                  raw_pkg_cls=None, pkg_cls=None, force=None, yield_none=False):
        """Generator that yields packages match a restriction.

        :type restrict: :obj:`pkgcore.restrictions.packages.PackageRestriction`
            instance.
        :param restrict: restriction to search via
        :param sorter: callable to do sorting during searching-
            if sorting the results, use this instead of sorting externally.
        :param pkg_filter: callable to do package filtering
        :param versioned: boolean controlling returning versioned or unversioned pkgs
        :param raw_pkg_cls: custom package class to use for generating raw pkg instances
        :param pkg_cls: custom package class to override raw pkg instances with
        :param yield_none: if True then itermatch will yield None for every
            non-matching package. This is meant for use in combination with
            C{twisted.task.cooperate} or other async uses where itermatch
            should not wait many (wallclock) seconds between yielding
            packages. If you override this method you should yield
            None in long-running loops, strictly calling it for every package
            is not necessary.
        """

        if not isinstance(restrict, restriction.base):
            raise TypeError(
                f"restrict must be a pkgcore.restriction.restrictions.base instance: "
                f"got {restrict!r}")

        if sorter is None:
            sorter = iter
        if pkg_filter is None:
            pkg_filter = iter
        if raw_pkg_cls is None:
            if versioned:
                raw_pkg_cls = self.package_class
            else:
                raw_pkg_cls = lambda *args: args

        if isinstance(restrict, atom):
            candidates = [(restrict.category, restrict.package)]
        else:
            candidates = self._identify_candidates(restrict, sorter)

        if force is None:
            match = restrict.match
        elif force:
            match = restrict.force_True
        else:
            match = restrict.force_False
        return self._internal_match(
            candidates, match, raw_pkg_cls=raw_pkg_cls, pkg_cls=pkg_cls,
            yield_none=yield_none, sorter=sorter, pkg_filter=pkg_filter,
            versioned=versioned)

    def _internal_gen_candidates(self, candidates, sorter, raw_pkg_cls, pkg_filter, versioned):
        for cp in sorter(candidates):
            if versioned:
                pkgs = (raw_pkg_cls(cp[0], cp[1], ver) for ver in self.versions.get(cp, ()))
            else:
                if self.versions.get(cp, ()):
                    pkgs = (raw_pkg_cls(cp[0], cp[1]),)
                else:
                    pkgs = ()
                pkgs = iter(pkgs)
            yield from sorter(pkg_filter(pkgs))

    def _internal_match(self, candidates, match_func, pkg_cls, yield_none=False, **kwargs):
        for pkg in self._internal_gen_candidates(candidates, **kwargs):
            if pkg_cls is not None:
                pkg = pkg_cls(pkg)

            if match_func(pkg):
                yield pkg
            elif yield_none:
                yield None

    def _identify_candidates(self, restrict, sorter):
        # full expansion
        if not isinstance(restrict, boolean.base) or isinstance(restrict, atom):
            return self._fast_identify_candidates(restrict, sorter)
        dsolutions = [
            ([c.restriction
              for c in collect_package_restrictions(x, ("category",))],
             [p.restriction
              for p in collect_package_restrictions(x, ("package",))])
            for x in restrict.iter_dnf_solutions(True)]

        # see if any solution state isn't dependent on cat/pkg in anyway.
        # if so, search whole search space.
        for x in dsolutions:
            if not x[0] and not x[1]:
                if sorter is iter:
                    return self.versions
                return (
                    (c, p)
                    for c in sorter(self.categories)
                    for p in sorter(self.packages.get(c, ())))

        # simple cases first.
        # if one specifies categories, and one doesn't
        cat_specified = bool(dsolutions[0][0])
        pkg_specified = bool(dsolutions[0][1])
        pgetter = self.packages.get
        if any(True for x in dsolutions[1:] if bool(x[0]) != cat_specified):
            if any(True for x in dsolutions[1:] if bool(x[1]) != pkg_specified):
                # merde.  so we've got a mix- some specify cats, some
                # don't, some specify pkgs, some don't.
                # this may be optimizable
                return self.versions
            # ok. so... one doesn't specify a category, but they all
            # specify packages (or don't)
            pr = values.OrRestriction(
                *tuple(iflatten_instance(
                    (x[1] for x in dsolutions if x[1]), values.base)))
            return (
                (c, p)
                for c in sorter(self.categories)
                for p in sorter(pgetter(c, [])) if pr.match(p))

        elif any(True for x in dsolutions[1:] if bool(x[1]) != pkg_specified):
            # one (or more) don't specify pkgs, but they all specify cats.
            cr = values.OrRestriction(
                *tuple(iflatten_instance(
                    (x[0] for x in dsolutions), values.base)))
            cats_iter = (c for c in sorter(self.categories) if cr.match(c))
            return (
                (c, p)
                for c in cats_iter for p in sorter(pgetter(c, [])))

        return self._fast_identify_candidates(restrict, sorter)

    def _fast_identify_candidates(self, restrict, sorter):
        pkg_restrict = set()
        cat_restrict = set()
        cat_exact = set()
        pkg_exact = set()

        for x in collect_package_restrictions(restrict,
                                              ("category", "package",)):
            if x.attr == "category":
                cat_restrict.add(x.restriction)
            elif x.attr == "package":
                pkg_restrict.add(x.restriction)

        for e, s in ((pkg_exact, pkg_restrict), (cat_exact, cat_restrict)):
            l = [x for x in s
                 if isinstance(x, values.StrExactMatch) and not x.negate]
            s.difference_update(l)
            e.update(x.exact for x in l)
        del l

        if restrict.negate:
            cat_exact = pkg_exact = ()

        if cat_exact:
            if not cat_restrict and len(cat_exact) == 1:
                # Cannot use pop here, cat_exact is reused below.
                c = next(iter(cat_exact))
                if not pkg_restrict and len(pkg_exact) == 1:
                    cp = (c, pkg_exact.pop())
                    if cp in self.versions:
                        return [cp]
                    return []
                cats_iter = [c]
            else:
                cat_restrict.add(values.ContainmentMatch(frozenset(cat_exact)))
                cats_iter = sorter(self._cat_filter(cat_restrict))
        elif cat_restrict:
            cats_iter = self._cat_filter(
                cat_restrict, negate=restrict.negate)
        else:
            cats_iter = sorter(self.categories)

        if pkg_exact:
            if not pkg_restrict:
                if sorter is iter:
                    pkg_exact = tuple(pkg_exact)
                else:
                    pkg_exact = sorter(pkg_exact)
                return (
                    (c, p)
                    for c in cats_iter for p in pkg_exact)
            else:
                pkg_restrict.add(values.ContainmentMatch(frozenset(pkg_exact)))

        if pkg_restrict:
            return self._package_filter(
                cats_iter, pkg_restrict, negate=restrict.negate)
        elif not cat_restrict:
            if sorter is iter and not cat_exact:
                return self.versions
            else:
                return (
                    (c, p) for c in
                    cats_iter for p in sorter(self.packages.get(c, ())))
        return (
            (c, p)
            for c in cats_iter for p in sorter(self.packages.get(c, ())))

    def _cat_filter(self, cat_restricts, negate=False):
        sentinel = not negate
        cats = [x.match for x in cat_restricts]
        for x in self.categories:
            for match in cats:
                if match(x) == sentinel:
                    yield x
                    break

    def _package_filter(self, cats_iter, pkg_restricts, negate=False):
        sentinel = not negate
        restricts = [x.match for x in pkg_restricts]
        pkgs_dict = self.packages
        for cat in cats_iter:
            for pkg in pkgs_dict.get(cat, ()):
                for match in restricts:
                    if match(pkg) == sentinel:
                        yield (cat, pkg)
                        break

    def notify_remove_package(self, pkg):
        """internal function

        notify the repository that a pkg it provides is being removed
        """
        ver_key = (pkg.category, pkg.package)
        l = [x for x in self.versions[ver_key] if x != pkg.fullver]
        if not l:
            # dead package
            wipe = list(self.packages[pkg.category]) == [pkg.package]
            self.packages.force_regen(pkg.category)
            if wipe:
                self.categories.force_regen(pkg.category)
        self.versions.force_regen(ver_key, tuple(l))

    def notify_add_package(self, pkg):
        """internal function

        notify the repository that a pkg is being added to it
        """
        ver_key = (pkg.category, pkg.package)
        s = set(self.versions.get(ver_key, ()))
        s.add(pkg.fullver)
        if pkg.category not in self.categories:
            self.categories.force_add(pkg.category)
        self.packages.force_regen(pkg.category)
        self.versions.force_regen(ver_key, tuple(s))

    @property
    def operations(self):
        return self.get_operations()

    def get_operations(self, observer=None):
        return self.operations_kls(self)

    def __bool__(self):
        try:
            next(iter(self.versions))
            return True
        except StopIteration:
            return False

    def __str__(self):
        if self.aliases:
            return str(self.aliases[0])
        return repr(self)

    @property
    def aliases(self):
        potentials = (getattr(self, key, None) for key in ('repo_id', 'location'))
        return tuple(x for x in potentials if x is not None)

    @jit_attr
    def masked(self):
        """Base package mask restriction."""
        return packages.OrRestriction(*self.pkg_masks)



"""Functions that turn a string into a restriction or raise ParseError.

__all__ = ("parse_match", "ParseError",)

@var parse_funcs: dict of the functions that are available.
"""

import re

from ..ebuild import atom, cpv, errors, restricts
from ..restrictions import packages, values
from ..restrictions.util import collect_package_restrictions

valid_globbing = re.compile(r"^(?:[\w+-.]+|(?<!\*)\*)+$").match


class ParseError(ValueError):
    """Raised if parsing a restriction expression failed."""


def comma_separated_containment(attr, values_kls=frozenset, token_kls=str):
    """Helper for parsing comma-separated strings to a ContainmentMatch.

    :param attr: name of the attribute.
    :return: a parse function: takes a string of comma-separated values,
        returns a :obj:`packages.PackageRestriction` matching packages that
        have any of those values in the attribute passed to this function.
    """
    def _parse(value):
        return packages.PackageRestriction(
            attr, values.ContainmentMatch(
                values_kls(token_kls(piece.strip()) for piece in value.split(','))
            )
        )
    return _parse


def convert_glob(token):
    if token in ('*', ''):
        return None
    elif '*' not in token:
        return values.StrExactMatch(token)
    elif not valid_globbing(token):
        raise ParseError(
            "globs must be composed of [\\w-.+], with optional "
            f"'*'- {token!r} is disallowed however")
    pattern = re.escape(token).replace('\\*', '.*')
    pattern = f"^{pattern}$"
    return values.StrRegex(pattern, match=True)


def collect_ops(text):
    i = 0
    while i < len(text) and text[i] in ("<", "=", ">", "~"):
        i += 1
    return text[0:i], text[i:]


def parse_match(text):
    """generate appropriate restriction for text

    Parsing basically breaks it down into chunks split by /, with each
    chunk allowing for prefix/postfix globbing- note that a postfixed
    glob on package token is treated as package attribute matching,
    not as necessarily a version match.

    If only one chunk is found, it's treated as a package chunk.
    Finally, it supports a nonstandard variation of atom syntax where
    the category can be dropped.

    Examples:

    - `*`: match all
    - `dev-*/*`: category must start with 'dev-'
    - `dev-*`: package must start with 'dev-'
    - `*-apps/portage*`: category must end in '-apps', package must start with
      'portage'
    - `>=portage-2.1`: atom syntax, package 'portage', version greater then or
      equal to '2.1'
    - dev-qt/*:5: all Qt 5 libs
    - boost:0/1.60: all packages named boost with a slot/subslot of 0/1.60.0

    :param text: string to attempt to parse
    :type text: string
    :return: :obj:`pkgcore.restrictions.packages` derivative
    """

    orig_text = text = text.strip()
    if "!" in text:
        raise ParseError(
            f"'!' or any form of blockers make no sense in this usage: {text!r}")

    restrictions = []
    if '::' in text:
        text, repo_id = text.rsplit('::', 1)
        restrictions.append(restricts.RepositoryDep(repo_id))
    if ':' in text:
        text, slot = text.rsplit(':', 1)
        slot, _sep, subslot = slot.partition('/')
        if slot:
            if '*' in slot:
                if r := convert_glob(slot):
                    restrictions.append(packages.PackageRestriction("slot", r))
            else:
                restrictions.append(restricts.SlotDep(slot))
        if subslot:
            if '*' in subslot:
                if r := convert_glob(subslot):
                    restrictions.append(packages.PackageRestriction("subslot", r))
            else:
                restrictions.append(restricts.SubSlotDep(subslot))

    tsplit = text.rsplit("/", 1)
    if len(tsplit) == 1:
        ops, text = collect_ops(text)
        if not ops:
            if "*" in text:
                if r := convert_glob(text):
                    restrictions.append(packages.PackageRestriction("package", r))
                else:
                    restrictions.append(packages.AlwaysTrue)
                if len(restrictions) == 1:
                    return restrictions[0]
                return packages.AndRestriction(*restrictions)
        elif text.startswith("*"):
            raise ParseError(
                f"cannot do prefix glob matches with version ops: {orig_text}")
        # ok... fake category.  whee.
        try:
            r = list(collect_package_restrictions(
                     atom.atom(f"{ops}category/{text}").restrictions,
                     attrs=("category",), invert=True))
        except errors.MalformedAtom as e:
            e.atom = orig_text
            raise ParseError(str(e)) from e
        if not restrictions and len(r) == 1:
            return r[0]
        restrictions.extend(r)
        return packages.AndRestriction(*restrictions)
    elif text[0] in atom.valid_ops or '*' not in text:
        # possibly a valid atom object
        try:
            return atom.atom(orig_text)
        except errors.MalformedAtom as e:
            if '*' not in text:
                raise ParseError(str(e)) from e
            # support globbed targets with version restrictions
            return packages.AndRestriction(*parse_globbed_version(text, orig_text))

    r = list(map(convert_glob, tsplit))
    if not r[0] and not r[1]:
        restrictions.append(packages.AlwaysTrue)
    elif not r[0]:
        restrictions.append(packages.PackageRestriction("package", r[1]))
    elif not r[1]:
        restrictions.append(packages.PackageRestriction("category", r[0]))
    else:
        restrictions.extend((
            packages.PackageRestriction("category", r[0]),
            packages.PackageRestriction("package", r[1]),
        ))
    if len(restrictions) == 1:
        return restrictions[0]
    return packages.AndRestriction(*restrictions)


def parse_globbed_version(text, orig_text):
    """Support parsing globbed targets with limited version restrictions.

    For example, '>=*/alsa-*-1.1.7' would match all packages named 'alsa-*'
    that are version 1.1.7 or greater.
    """
    restrictions = []
    # find longest matching op
    op = max(x for x in atom.valid_ops if text.startswith(x))
    text = text[len(op):]
    # determine pkg version
    chunks = text.rsplit('-', 1)
    if len(chunks) == 1:
        raise ParseError(f'missing valid package version: {orig_text!r}')
    version_txt = chunks[-1]
    version = cpv.isvalid_version_re.match(version_txt)
    if not version:
        if '*' in version_txt:
            raise ParseError(
                f'operator {op!r} invalid with globbed version: {version_txt!r}')
        raise ParseError(f'missing valid package version: {orig_text!r}')
    restrictions.append(restricts.VersionMatch(op, version.group(0)))
    # parse the remaining chunk
    restrictions.append(parse_match(chunks[0]))
    return restrictions


def parse_pv(repo, text):
    """Return a CPV instance from either a cpv or a pv string.

    If a pv is passed it needs to match a single cpv in repo.
    """
    try:
        return cpv.CPV.versioned(text)
    except errors.InvalidCPV:
        restrict = parse_match(f"={text}")
        result = None
        for match in repo.itermatch(restrict):
            if result is not None:
                raise ParseError(
                    f"multiple matches for {text} ({result.cpvstr}, {match.cpvstr})")
            result = match
        if result is None:
            raise ParseError(f"no matches for {text}")
        return cpv.CPV(result.category, result.package, result.version)


parse_funcs = {
    'match': parse_match,
}



# "More than one statement on a single line"
# pylint: disable-msg=C0321

"""
atom exceptions
"""

__all__ = ("MalformedAtom", "InvalidVersion", "InvalidCPV", "DepsetParseError")

import textwrap

from ..exceptions import PkgcoreException
from ..package import errors


class MalformedAtom(errors.InvalidDependency):
    """Package atom doesn't follow required specifications."""

    def __init__(self, atom, err=None):
        self.atom = atom
        self.err = err
        super().__init__(str(self))

    def __str__(self):
        msg = f'invalid package atom: {self.atom!r}'
        if self.err:
            msg += f': {self.err}'
        return msg


class InvalidVersion(errors.InvalidDependency):
    """Package version doesn't follow required specifications."""

    def __init__(self, ver, rev, err=''):
        super().__init__(
            f"Version restriction ver='{ver}', rev='{rev}', is malformed: error {err}")
        self.ver, self.rev, self.err = ver, rev, err


class InvalidCPV(errors.InvalidPackageName):
    """CPV with unsupported characters or format."""


class DepsetParseError(errors.InvalidDependency):

    def __init__(self, s, token=None, msg=None, attr=None):
        self.dep_str = s
        self.token = token
        self.msg = msg
        self.attr = attr

    def __str__(self):
        msg = []
        if self.attr is not None:
            msg.append(f'failed parsing {self.attr}')
        msg.append(f'{self.dep_str!r} is unparseable')
        if self.token is not None:
            msg.append(f'flagged token- {self.token}')
        if self.msg is not None:
            msg.append(f'{self.msg}')
        return ': '.join(msg)


class SanityCheckError(PkgcoreException):
    """Generic error for sanity check failures."""

    def msg(self, verbosity, prefix='  '):
        if verbosity > 0:
            return self.verbose_msg(prefix)
        else:
            return f'{prefix}{self}'


class PkgPretendError(SanityCheckError):
    """The pkg_pretend phase check failed for a package."""

    def __init__(self, pkg, output, error):
        self.pkg = pkg
        self.output = output
        self.error = error

    def msg(self, verbosity=0, prefix=' '):
        header = f'>>> {self.pkg.cpvstr}: failed pkg_pretend'
        msg = []
        error_msg = self.error.msg(verbosity=verbosity)
        if verbosity > 0:
            msg.extend(self.output.splitlines())
            msg.extend(error_msg.splitlines())
            msg = [f'{prefix}{l}' for l in msg]
        elif error_msg:
            header += f': {error_msg}'
        return '\n'.join([header] + msg)


class RequiredUseError(SanityCheckError):
    """REQUIRED_USE check(s) for a package failed."""

    def __init__(self, pkg, unmatched):
        self.pkg = pkg
        self.unmatched = unmatched

    def msg(self, verbosity=0, prefix='  '):
        header = f'>>> {self.pkg.cpvstr}: failed REQUIRED_USE'
        errors = []
        for node in self.unmatched:
            errors.append(textwrap.dedent(
                f"""
                Failed to match: {node}
                from: {self.pkg.required_use}
                for USE: {' '.join(sorted(self.pkg.use))}
                """
            ))
        msg = [f'{prefix}{line}' for e in errors for line in e.strip().splitlines()]
        return '\n'.join([header] + msg)



import logging
import os
import shutil
import sys
import tempfile
from unittest import mock

from pkgcore import plugin
from pkgcore.test import silence_logging
from snakeoil.osutils import pjoin
from snakeoil.sequences import stable_unique


class LowPlug:
    priority = 1


class TestModules:

    def setup_method(self, method):
        self.dir = tempfile.mkdtemp()
        self.dir2 = tempfile.mkdtemp()

        # force plugin module to use package dir for cache dir by setting
        # system/user cache dirs to nonexistent paths
        self.patcher = mock.patch('pkgcore.plugin.const')
        const = self.patcher.start()
        const.SYSTEM_CACHE_PATH = pjoin(self.dir, 'nonexistent')
        const.USER_CACHE_PATH = pjoin(self.dir, 'nonexistent')

        # Set up some test modules for our use.
        self.packdir = pjoin(self.dir, 'mod_testplug')
        self.packdir2 = pjoin(self.dir2, 'mod_testplug')
        os.mkdir(self.packdir)
        os.mkdir(self.packdir2)
        with open(pjoin(self.packdir, '__init__.py'), 'w') as init:
            init.write('''
from pkgcore.plugins import extend_path

extend_path(__path__, __name__)
''')
        filename = pjoin(self.packdir, 'plug.py')
        with open(filename, 'w') as plug:
            plug.write('''
class DisabledPlug:
    disabled = True

class HighPlug:
    priority = 7

class LowPlug:
    priority = 1

low_plug = LowPlug()
high_plug = HighPlug()

pkgcore_plugins = {
    'plugtest': [
        DisabledPlug,
        high_plug,
        'tests.test_plugin.LowPlug',
    ]
}
''')
        # Move the mtime 2 seconds into the past so the .pyc file has
        # a different mtime.
        st = os.stat(filename)
        os.utime(filename, (st.st_atime, st.st_mtime - 2))
        with open(pjoin(self.packdir, 'plug2.py'), 'w') as plug2:
            plug2.write('# I do not have any pkgcore_plugins for you!\n')
        with open(pjoin(self.packdir2, 'plug.py'), 'w') as plug:
            plug.write('''
# This file is later on sys.path than the plug.py in packdir, so it should
# not have any effect on the tests.

class HiddenPlug:
    priority = 8

pkgcore_plugins = {'plugtest': [HiddenPlug]}
''')
        # Append it to the path
        sys.path.insert(0, self.dir2)
        sys.path.insert(0, self.dir)

    def teardown_method(self):
        # stop mocked patcher
        self.patcher.stop()
        # pop the test module dir from path
        sys.path.pop(0)
        sys.path.pop(0)
        # and kill it
        shutil.rmtree(self.dir)
        shutil.rmtree(self.dir2)
        # make sure we don't keep the sys.modules entries around
        sys.modules.pop('mod_testplug', None)
        sys.modules.pop('mod_testplug.plug', None)
        sys.modules.pop('mod_testplug.plug2', None)

    def test_extend_path(self):
        import mod_testplug
        expected = stable_unique(
            pjoin(p, 'mod_testplug')
            for p in sys.path if os.path.isdir(p))
        assert expected == mod_testplug.__path__, \
            set(expected) ^ set(mod_testplug.__path__)

    def _runit(self, method):
        plugin._global_cache.clear()
        method()
        mtime = os.path.getmtime(pjoin(self.packdir, plugin.CACHE_FILENAME))
        method()
        plugin._global_cache.clear()
        method()
        method()
        assert mtime == \
            os.path.getmtime(pjoin(self.packdir, plugin.CACHE_FILENAME))
        # We cannot write this since it contains an unimportable plugin.
        assert not os.path.exists(pjoin(self.packdir2, plugin.CACHE_FILENAME))

    def _test_plug(self):
        import mod_testplug
        assert plugin.get_plugin('spork', mod_testplug) is None
        plugins = list(plugin.get_plugins('plugtest', mod_testplug))
        assert len(plugins) == 2, plugins
        plugin.get_plugin('plugtest', mod_testplug)
        assert 'HighPlug' == \
            plugin.get_plugin('plugtest', mod_testplug).__class__.__name__
        with open(pjoin(self.packdir, plugin.CACHE_FILENAME)) as f:
            lines = f.readlines()
        assert len(lines) == 3
        assert plugin.CACHE_HEADER + "\n" == lines[0]
        lines.pop(0)
        lines.sort()
        mtime = int(os.path.getmtime(pjoin(self.packdir, 'plug2.py')))
        assert f'plug2:{mtime}:\n' == lines[0]
        mtime = int(os.path.getmtime(pjoin(self.packdir, 'plug.py')))
        assert (
            f'plug:{mtime}:plugtest,7,1:plugtest,1,tests.test_plugin.LowPlug:plugtest,0,0\n'
            == lines[1])

    def test_plug(self):
        self._runit(self._test_plug)

    def _test_no_unneeded_import(self):
        import mod_testplug
        list(plugin.get_plugins('spork', mod_testplug))
        sys.modules.pop('mod_testplug.plug')
        # This one is not loaded if we are testing with a good cache.
        sys.modules.pop('mod_testplug.plug2', None)
        list(plugin.get_plugins('plugtest', mod_testplug))
        # Extra messages since getting all of sys.modules printed is annoying.
        assert 'mod_testplug.plug' in sys.modules, 'plug not loaded'
        assert 'mod_testplug.plug2' not in sys.modules, 'plug2 loaded'

    def test_no_unneeded_import(self):
        self._runit(self._test_no_unneeded_import)

    @silence_logging(logging.root)
    def test_cache_corruption(self):
        print(plugin.const)
        print('wheeeeee')
        import mod_testplug
        list(plugin.get_plugins('spork', mod_testplug))
        filename = pjoin(self.packdir, plugin.CACHE_FILENAME)
        cachefile = open(filename, 'a')
        try:
            cachefile.write('corruption\n')
        finally:
            cachefile.close()
        # Shift the file into the past a little or the rewritten file
        # will occasionally have the same mtime as the corrupt one.
        st = os.stat(filename)
        corrupt_mtime = st.st_mtime - 2
        os.utime(filename, (st.st_atime, corrupt_mtime))
        plugin._global_cache.clear()
        self._test_plug()
        good_mtime = os.path.getmtime(
            pjoin(self.packdir, plugin.CACHE_FILENAME))
        plugin._global_cache.clear()
        self._test_plug()
        assert good_mtime == os.path.getmtime(pjoin(self.packdir, plugin.CACHE_FILENAME))
        assert good_mtime != corrupt_mtime

    def test_rewrite_on_remove(self):
        filename = pjoin(self.packdir, 'extra.py')
        plug = open(filename, 'w')
        try:
            plug.write('pkgcore_plugins = {"plugtest": [object()]}\n')
        finally:
            plug.close()

        plugin._global_cache.clear()
        import mod_testplug
        assert len(list(plugin.get_plugins('plugtest', mod_testplug))) == 3

        os.unlink(filename)

        plugin._global_cache.clear()
        self._test_plug()

    @silence_logging(logging.root)
    def test_priority_caching(self):
        plug3 = open(pjoin(self.packdir, 'plug3.py'), 'w')
        try:
            plug3.write('''
class LowPlug:
    priority = 6

pkgcore_plugins = {
    'plugtest': [LowPlug()],
}
''')
        finally:
            plug3.close()
        plug4 = open(pjoin(self.packdir, 'plug4.py'), 'w')
        try:
            plug4.write('''
# First file tried, only a disabled plugin.
class HighDisabledPlug:
    priority = 15
    disabled = True

pkgcore_plugins = {
    'plugtest': [HighDisabledPlug()],
}
''')
        finally:
            plug4.close()
        plug5 = open(pjoin(self.packdir, 'plug5.py'), 'w')
        try:
            plug5.write('''
# Second file tried, with a skipped low priority plugin.
class HighDisabledPlug:
    priority = 12
    disabled = True

class LowPlug:
    priority = 6

pkgcore_plugins = {
    'plugtest': [HighDisabledPlug(), LowPlug()],
}
''')
        finally:
            plug5.close()
        plug6 = open(pjoin(self.packdir, 'plug6.py'), 'w')
        try:
            plug6.write('''
# Not tried, bogus priority.
class BogusPlug:
    priority = 'spoon'

pkgcore_plugins = {
    'plugtest': [BogusPlug()],
}
''')
        finally:
            plug6.close()
        self._runit(self._test_priority_caching)

    def _test_priority_caching(self):
        import mod_testplug
        list(plugin.get_plugins('spork', mod_testplug))
        sys.modules.pop('mod_testplug.plug', None)
        sys.modules.pop('mod_testplug.plug2', None)
        sys.modules.pop('mod_testplug.plug3', None)
        sys.modules.pop('mod_testplug.plug4', None)
        sys.modules.pop('mod_testplug.plug5', None)
        sys.modules.pop('mod_testplug.plug6', None)
        best_plug = plugin.get_plugin('plugtest', mod_testplug)
        from mod_testplug import plug
        assert plug.high_plug == best_plug
        # Extra messages since getting all of sys.modules printed is annoying.
        assert 'mod_testplug.plug' in sys.modules, 'plug not loaded'
        assert 'mod_testplug.plug2' not in sys.modules, 'plug2 loaded'
        assert 'mod_testplug.plug3' not in sys.modules, 'plug3 loaded'
        assert 'mod_testplug.plug4' in sys.modules, 'plug4 not loaded'
        assert 'mod_testplug.plug5' in sys.modules, 'plug4 not loaded'
        assert 'mod_testplug.plug6' not in sys.modules, 'plug6 loaded'

    @silence_logging(logging.root)
    def test_header_change_invalidates_cache(self):
        # Write the cache
        plugin._global_cache.clear()
        import mod_testplug
        list(plugin.get_plugins('testplug', mod_testplug))

        # Modify the cache.
        filename = pjoin(self.packdir, plugin.CACHE_FILENAME)
        with open(filename) as f:
            cache = f.readlines()
        cache[0] = 'not really a pkgcore plugin cache\n'
        with open(filename, 'w') as f:
            f.write(''.join(cache))

        # And test if it is properly rewritten.
        plugin._global_cache.clear()
        self._test_plug()



"""
build operation
"""

__all__ = (
    'build_base', 'install', 'uninstall', 'replace', 'fetch_base',
    'empty_build_op', 'FailedDirectory', 'GenericBuildError',
)

import os

from snakeoil import klass
from snakeoil.dependant_methods import ForcedDepends
from snakeoil.osutils import pjoin

from .. import operations as _operations_mod
from ..exceptions import PkgcoreUserException
from ..fetch import custom as fetch_custom
from ..fetch import errors as fetch_errors


class fetch_base:

    def __init__(self, domain, pkg, fetchables, distdir=None):
        self.verified_files = {}
        self._basenames = set()
        self.domain = domain
        self.pkg = pkg
        self.fetchables = fetchables
        self.distdir = distdir if distdir is not None else domain.distdir

        # create fetcher
        fetchcmd = domain.settings['FETCHCOMMAND']
        resumecmd = domain.settings.get('RESUMECOMMAND', fetchcmd)
        attempts = int(domain.settings.get('FETCH_ATTEMPTS', 10))
        self.fetcher = fetch_custom.fetcher(
            self.distdir, fetchcmd, resumecmd, attempts=attempts)

    def fetch_all(self, observer):
        # TODO: add parallel fetch support
        failures = []
        for fetchable in self.fetchables:
            if not self.fetch_one(fetchable, observer):
                failures.append(fetchable)
        return self.verified_files, failures

    def fetch_one(self, fetchable, observer, retry=False):
        if fetchable.filename in self._basenames:
            return True
        # fetching files without uri won't fly
        # XXX hack atm, could use better logic but works for now
        try:
            fp = self.fetcher(fetchable)
        except fetch_errors.ChksumFailure as e:
            # checksum failed, rename file and try refetching
            path = pjoin(self.distdir, fetchable.filename)
            failed_filename = f'{fetchable.filename}._failed_chksum_'
            failed_path = pjoin(self.distdir, failed_filename)
            os.rename(path, failed_path)
            if retry:
                raise
            observer.error(str(e))
            observer.error(f'renaming to {failed_filename!r} and refetching from upstream')
            observer.flush()
            # refetch directly from upstream
            return self.fetch_one(fetchable.upstream, observer, retry=True)
        except fetch_errors.FetchFailed:
            fp = None
        if fp is None:
            return False
        self.verified_files[fp] = fetchable
        self._basenames.add(fetchable.filename)
        return True


class operations(_operations_mod.base):

    _fetch_kls = fetch_base

    def __init__(self, domain, pkg, observer=None, disable_overrides=(),
                 enable_overrides=()):
        self.observer = observer
        self.pkg = pkg
        self.domain = domain
        self.verified_files = None
        super().__init__(disable_overrides, enable_overrides)

    def _cmd_api_info(self):
        return self._cmd_implementation_info()

    @_operations_mod.is_standalone
    def _cmd_api_mergable(self):
        return getattr(self.pkg, 'built', False)

    def _cmd_api_sanity_check(self):
        return self._cmd_implementation_sanity_check(self.domain)

    def _cmd_implementation_sanity_check(self, domain):
        return True

    def _cmd_api_localize(self, force=False, observer=klass.sentinel):
        observer = observer if observer is not klass.sentinel else self.observer
        return self._cmd_implementation_localize(
            self._get_observer(observer), force=force)

    def _cmd_api_cleanup(self, force=False, observer=klass.sentinel):
        observer = observer if observer is not klass.sentinel else self.observer
        return self._cmd_implementation_cleanup(
            self._get_observer(observer), force=force)

    def _cmd_api_configure(self, observer=klass.sentinel):
        observer = observer if observer is not klass.sentinel else self.observer
        return self._cmd_implementation_configure(
            self._get_observer(observer))

    @_operations_mod.is_standalone
    def _cmd_api_fetch(self, fetchables=None, observer=klass.sentinel, distdir=None):
        observer = observer if observer is not klass.sentinel else self.observer
        if fetchables is None:
            fetchables = self.pkg.fetchables
        elif not isinstance(fetchables, (tuple, list)):
            fetchables = [fetchables]
        fetcher = self._fetch_kls(self.domain, self.pkg, fetchables, distdir)
        verified, failures = fetcher.fetch_all(self._get_observer(observer))

        if failures:
            # run pkg_nofetch phase for fetch restricted pkgs
            if 'fetch' in self.pkg.restrict:
                # This requires wrapped packages from a configured repo, otherwise
                # buildables aren't available to run the pkg_nofetch phase.
                configured_repo = self.domain.unfiltered_repos[self.pkg.repo.repo_id]
                pkgwrap = configured_repo.package_class(self.pkg)
                build_ops = self.domain.build_pkg(pkgwrap, observer, failed=True)
                build_ops.nofetch()
                build_ops.cleanup(force=True)
            observer.error('failed fetching files: %s::%s', self.pkg.cpvstr, self.pkg.repo.repo_id)
            raise GenericBuildError('failed fetching required distfiles')

        self.verified_files = verified
        return True


class build_operations(operations):

    __required__ = frozenset(["build"])

    def _cmd_api_build(self, observer=None, failed=False, clean=True, **kwargs):
        return self._cmd_implementation_build(
            self._get_observer(observer),
            self.verified_files,
            clean=clean, **kwargs)

    def _cmd_api_buildable(self, domain):
        return self._cmd_implementation_buildable(domain)

    def _cmd_implementation_buildable(self, domain):
        return True


class build_base(metaclass=ForcedDepends):

    stage_depends = {'finish': 'start'}

    def __init__(self, domain, observer):
        self.domain = domain
        self.observer = observer

    def start(self):
        return True

    def finish(self):
        return True


class build(build_base):

    stage_depends = {
        "setup": "start",
        "unpack": "setup",
        "configure": "prepare",
        "prepare": "unpack",
        "compile": "configure",
        "test": "compile",
        "install": "test",
        "finalize": "install",
    }

    def __init__(self, domain, pkg, verified_files, observer):
        super().__init__(domain, observer)
        self.pkg = pkg
        self.verified_files = verified_files

    def setup(self):
        return True

    def unpack(self):
        return True

    def prepare(self):
        return True

    def configure(self):
        return True

    def compile(self):
        return True

    def test(self):
        return True

    def install(self):
        return True

    def finalize(self):
        """finalize any build steps required"""
        return True

    def cleanup(self):
        """cleanup any working files/dirs created during building"""
        return True

    for k in ("setup", "unpack", "configure", "compile", "test", "install"):
        locals()[k].__doc__ = (
            "execute any %s steps required; "
            "implementations of this interface should overide this as needed"
            % k)
    for k in ("setup", "unpack", "configure", "compile", "test", "install", "finalize"):
        o = locals()[k]
        o.__doc__ = "\n".join(x.lstrip() for x in o.__doc__.split("\n") + [
                              ":return: True on success, False on failure"])
    del o, k


class install(build_base):

    stage_depends = {
        "preinst": "start",
        "postinst": "preinst",
        "finalize": "postinst",
    }

    def __init__(self, domain, newpkg, observer):
        super().__init__(domain, observer)
        self.new_pkg = self.pkg = newpkg

    def add_triggers(self, engine):
        pass

    def preinst(self):
        """any pre merge steps needed"""
        return True

    def postinst(self):
        """any post merge steps needed"""
        return True

    def finalize(self):
        """finalize any merge steps required"""
        return True


class uninstall(build_base):

    stage_depends = {
        "prerm": "start",
        "postrm": "prerm",
        "finalize": "postrm",
    }

    def __init__(self, domain, oldpkg, observer):
        super().__init__(domain, observer)
        self.old_pkg = self.pkg = oldpkg

    def add_triggers(self, engine):
        pass

    def prerm(self):
        """any pre unmerge steps needed"""
        return True

    def postrm(self):
        """any post unmerge steps needed"""
        return True

    def postinst(self):
        """any post unmerge steps needed"""
        return True

    def finalize(self):
        """finalize any unmerge steps required"""
        return True


class replace(install, uninstall):

    stage_depends = {
        "finalize": "postinst",
        "postinst": "postrm",
        "postrm": "prerm",
        "prerm": "preinst",
        "preinst": "start",
    }

    def __init__(self, domain, old_pkg, new_pkg, observer):
        build_base.__init__(self, domain, observer)
        self.new_pkg = new_pkg
        self.old_pkg = old_pkg


class empty_build_op(build_base):

    stage_depends = {}

    def __init__(self, pkg, observer=None, clean=False):
        super().__init__(observer)
        self.pkg = pkg

    def cleanup(self):
        return True

    def finalize(self):
        return self.pkg


class BuildError(PkgcoreUserException):
    pass


class FailedDirectory(BuildError):
    def __init__(self, path, text):
        super().__init__(f"failed creating/ensuring dir {path}: {text}")


class GenericBuildError(BuildError):
    def __init__(self, err):
        super().__init__(f"failed build operation: {err}")
        self.err = str(err)
